from collections import defaultdict

from cadnano2.model.document import Document
from cadnano2.model.enum import LatticeType, StrandType
from cadnano2.model.parts.honeycombpart import HoneycombPart
from cadnano2.model.parts.squarepart import SquarePart
from cadnano2.model.virtualhelix import VirtualHelix
from cadnano2.views import styles
import cadnano2.util as util
import cadnano2.cadnano as cadnano
# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtGui', globals(),  ['QColor'])
if cadnano.app().isGui():
    from cadnano2.ui.dialogs.ui_latticetype import Ui_LatticeType
    util.qtWrapImport('QtWidgets', globals(),  ['QDialog', 'QDialogButtonBox'])

NODETAG = "node"
NAME = "name"
OBJ_ID = "objectid"
INST_ID = "instanceid"
DONE = "done"
CHECKED = "check"
LOCKED = "locked"

VHELIX = "vhelix"
NUM = "num"
COL = "col"
ROW = "row"
SCAFFOLD = "scaffold"
STAPLE = "staple"
INSERTION = "insertion"
DELETION = "deletion"


def import_legacy_dict(document, obj, latticeType=LatticeType.Honeycomb):
    """
    Parses a dictionary (obj) created from reading a json file and uses it
    to populate the given document with model data.
    """
    numBases = len(obj['vstrands'][0]['scaf'])
    if cadnano.app().isGui():
        # from ui.dialogs.ui_latticetype import Ui_LatticeType
        # util.qtWrapImport('QtGui', globals(),  ['QDialog', 'QDialogButtonBox'])
        dialog = QDialog()
        dialogLT = Ui_LatticeType()
        dialogLT.setupUi(dialog)
        # DETERMINE LATTICE TYPE
        if numBases % 21 == 0 and numBases % 32 == 0:
            if dialog.exec() == 1:
                latticeType = LatticeType.Square
            else:
                latticeType = LatticeType.Honeycomb
        elif numBases % 32 == 0:
            latticeType = LatticeType.Square
        elif numBases % 21 == 0:
            latticeType = LatticeType.Honeycomb
        else:
            if dialog.exec() == 1:
                latticeType = LatticeType.Square
            else:
                latticeType = LatticeType.Honeycomb
    else:  # Headless, assume the latticeType arg was meaningful
        pass

    # DETERMINE MAX ROW,COL
    maxRowJson = maxColJson = 0
    for helix in obj['vstrands']:
        maxRowJson = max(maxRowJson, int(helix['row'])+1)
        maxColJson = max(maxColJson, int(helix['col'])+1)

    # CREATE PART ACCORDING TO LATTICE TYPE
    if latticeType == LatticeType.Honeycomb:
        steps = numBases/21
        nRows = max(30, maxRowJson, cadnano.app().prefs.honeycombRows)
        nCols = max(32, maxColJson, cadnano.app().prefs.honeycombCols)
        part = HoneycombPart(document=document, maxRow=nRows, maxCol=nCols, maxSteps=steps)
    elif latticeType == LatticeType.Square:
        isSQ100 = True  # check for custom SQ100 format
        for helix in obj['vstrands']:
            if helix['col'] != 0:
                isSQ100 = False
                break
        if isSQ100:
            dialogLT.label.setText("Is this a SQ100 file?")
            if dialog.exec() == 1:
                nRows, nCols = 100, 1
            else:
                nRows, nCols = 40, 30
        else:
            nRows, nCols = 40, 30
        steps = numBases/32
        nRows = max(30, maxRowJson, cadnano.app().prefs.squareRows)
        nCols = max(32, maxColJson, cadnano.app().prefs.squareCols)
        part = SquarePart(document=document, maxRow=nRows, maxCol=nCols, maxSteps=steps)
    else:
        raise TypeError("Lattice type not recognized")
    document._addPart(part, useUndoStack=False)

    # POPULATE VIRTUAL HELICES
    orderedCoordList = []
    vhNumToCoord = {}
    for helix in obj['vstrands']:
        vhNum = helix['num']
        row = helix['row']
        col = helix['col']
        scaf= helix['scaf']
        coord = (row, col)
        vhNumToCoord[vhNum] = coord
        orderedCoordList.append(coord)
    # make sure we retain the original order
    for vhNum in sorted(vhNumToCoord.keys()):
        row, col = vhNumToCoord[vhNum]
        part.createVirtualHelix(row, col, useUndoStack=False)
    part.setImportedVHelixOrder(orderedCoordList)

    # INSTALL STRANDS AND COLLECT XOVER LOCATIONS
    numHelixes = len(obj['vstrands'])-1
    scaf_seg = defaultdict(list)
    scaf_xo = defaultdict(list)
    stap_seg = defaultdict(list)
    stap_xo = defaultdict(list)
    try:
        for helix in obj['vstrands']:
            vhNum = helix['num']
            row = helix['row']
            col = helix['col']
            scaf = helix['scaf']
            stap = helix['stap']
            insertions = helix['loop']
            skips = helix['skip']
            vh = part.virtualHelixAtCoord((row, col))
            scafStrandSet = vh.scaffoldStrandSet()
            stapStrandSet = vh.stapleStrandSet()
            assert(len(scaf)==len(stap) and len(stap)==part.maxBaseIdx()+1 and\
                   len(scaf)==len(insertions) and len(insertions)==len(skips))
            # read scaffold segments and xovers
            for i in range(len(scaf)):
                fiveVH, fiveIdx, threeVH, threeIdx = scaf[i]
                if fiveVH == -1 and threeVH == -1:
                    continue  # null base
                if isSegmentStartOrEnd(StrandType.Scaffold, vhNum, i, fiveVH,\
                                       fiveIdx, threeVH, threeIdx):
                    scaf_seg[vhNum].append(i)
                if fiveVH != vhNum and threeVH != vhNum:  # special case
                    scaf_seg[vhNum].append(i)  # end segment on a double crossover
                if is3primeXover(StrandType.Scaffold, vhNum, i, threeVH, threeIdx):
                    scaf_xo[vhNum].append((i, threeVH, threeIdx))
            assert (len(scaf_seg[vhNum]) % 2 == 0)
            # install scaffold segments
            for i in range(0, len(scaf_seg[vhNum]), 2):
                lowIdx = scaf_seg[vhNum][i]
                highIdx = scaf_seg[vhNum][i+1]
                scafStrandSet.createStrand(lowIdx, highIdx, useUndoStack=False)
            # read staple segments and xovers
            for i in range(len(stap)):
                fiveVH, fiveIdx, threeVH, threeIdx = stap[i]
                if fiveVH == -1 and threeVH == -1:
                    continue  # null base
                if isSegmentStartOrEnd(StrandType.Staple, vhNum, i, fiveVH,\
                                       fiveIdx, threeVH, threeIdx):
                    stap_seg[vhNum].append(i)
                if fiveVH != vhNum and threeVH != vhNum:  # special case
                    stap_seg[vhNum].append(i)  # end segment on a double crossover
                if is3primeXover(StrandType.Staple, vhNum, i, threeVH, threeIdx):
                    stap_xo[vhNum].append((i, threeVH, threeIdx))
            assert (len(stap_seg[vhNum]) % 2 == 0)
            # install staple segments
            for i in range(0, len(stap_seg[vhNum]), 2):
                lowIdx = stap_seg[vhNum][i]
                highIdx = stap_seg[vhNum][i+1]
                stapStrandSet.createStrand(lowIdx, highIdx, useUndoStack=False)
    except AssertionError:
        if not cadnano.app().isGui():
            print("Unrecognized file format.")
        else:
            dialogLT.label.setText("Unrecognized file format.")
            dialogLT.buttonBox.setStandardButtons(QDialogButtonBox.Ok)
            dialog.exec()

    # INSTALL XOVERS
    for helix in obj['vstrands']:
        vhNum = helix['num']
        row = helix['row']
        col = helix['col']
        scaf = helix['scaf']
        stap = helix['stap']
        insertions = helix['loop']
        skips = helix['skip']
        fromVh = part.virtualHelixAtCoord((row, col))
        scafStrandSet = fromVh.scaffoldStrandSet()
        stapStrandSet = fromVh.stapleStrandSet()
        # install scaffold xovers
        for (idx5p, toVhNum, idx3p) in scaf_xo[vhNum]:
            # idx3p is 3' end of strand5p, idx5p is 5' end of strand3p
            strand5p = scafStrandSet.getStrand(idx5p)
            toVh = part.virtualHelixAtCoord(vhNumToCoord[toVhNum])
            strand3p = toVh.scaffoldStrandSet().getStrand(idx3p)
            part.createXover(strand5p, idx5p, strand3p, idx3p, useUndoStack=False)
        # install staple xovers
        for (idx5p, toVhNum, idx3p) in stap_xo[vhNum]:
            # idx3p is 3' end of strand5p, idx5p is 5' end of strand3p
            strand5p = stapStrandSet.getStrand(idx5p)
            toVh = part.virtualHelixAtCoord(vhNumToCoord[toVhNum])
            strand3p = toVh.stapleStrandSet().getStrand(idx3p)
            part.createXover(strand5p, idx5p, strand3p, idx3p, useUndoStack=False)

    # SET DEFAULT COLOR
    for oligo in part.oligos():
        if oligo.isStaple():
            defaultColor = styles.DEFAULT_STAP_COLOR
        else:
            defaultColor = styles.DEFAULT_SCAF_COLOR
        oligo.applyColor(defaultColor, useUndoStack=False)

    # COLORS, INSERTIONS, SKIPS
    for helix in obj['vstrands']:
        vhNum = helix['num']
        row = helix['row']
        col = helix['col']
        scaf = helix['scaf']
        stap = helix['stap']
        insertions = helix['loop']
        skips = helix['skip']
        vh = part.virtualHelixAtCoord((row, col))
        scafStrandSet = vh.scaffoldStrandSet()
        stapStrandSet = vh.stapleStrandSet()
        # install insertions and skips
        for baseIdx in range(len(stap)):
            sumOfInsertSkip = insertions[baseIdx] + skips[baseIdx]
            if sumOfInsertSkip != 0:
                scaf_strand = scafStrandSet.getStrand(baseIdx)
                stap_strand = stapStrandSet.getStrand(baseIdx)
                if scaf_strand:
                    scaf_strand.addInsertion(baseIdx, sumOfInsertSkip, useUndoStack=False)
                elif stap_strand:
                    stap_strand.addInsertion(baseIdx, sumOfInsertSkip, useUndoStack=False)
        # end for
        # populate colors
        for baseIdx, colorNumber in helix['stap_colors']:
            color = QColor((colorNumber>>16)&0xFF, (colorNumber>>8)&0xFF, colorNumber&0xFF).name()
            strand = stapStrandSet.getStrand(baseIdx)
            strand.oligo().applyColor(color, useUndoStack=False)

def isSegmentStartOrEnd(strandType, vhNum, baseIdx, fiveVH, fiveIdx, threeVH, threeIdx):
    """Returns True if the base is a breakpoint or crossover."""
    if strandType == StrandType.Scaffold:
        offset = 1
    else:
        offset = -1
    if (fiveVH == vhNum and threeVH != vhNum):
        return True
    if (fiveVH != vhNum and threeVH == vhNum):
        return True
    if (vhNum % 2 == 0 and fiveVH == vhNum and fiveIdx != baseIdx-offset):
        return True
    if (vhNum % 2 == 0 and threeVH == vhNum and threeIdx != baseIdx+offset):
        return True
    if (vhNum % 2 == 1 and fiveVH == vhNum and fiveIdx != baseIdx+offset):
        return True
    if (vhNum % 2 == 1 and threeVH == vhNum and threeIdx != baseIdx-offset):
        return True
    if (fiveVH == -1 and threeVH != -1):
        return True
    if (fiveVH != -1 and threeVH == -1):
        return True
    return False

def is3primeXover(strandType, vhNum, baseIdx, threeVH, threeIdx):
    """Returns True of the threeVH doesn't match vhNum, or threeIdx
    is not a natural neighbor of baseIdx."""
    if threeVH == -1:
        return False
    if vhNum != threeVH:
        return True
    if strandType == StrandType.Scaffold:
        offset = 1
    else:
        offset = -1
    if (vhNum % 2 == 0 and threeVH == vhNum and threeIdx != baseIdx+offset):
        return True
    if (vhNum % 2 == 1 and threeVH == vhNum and threeIdx != baseIdx-offset):
        return True
    return False



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='cadnano2',
    version='2.4.7',    
    description='Cadnano2 for PyQt6',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/douglaslab/cadnano2',
    author='Shawn Douglas',
    author_email='shawn.douglas@ucsf.edu',
    license='MIT',
    packages=find_packages(),
    package_data={'cadnano2': ['ui/mainwindow/images/*.svg', 'ui/mainwindow/images/*.png']},
    install_requires=['PyQt6'],
    entry_points = {'console_scripts': ['cadnano2 = cadnano2.main:main']},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
    ],
)




from ..ui.dialogs.ui_preferences import Ui_Preferences
from ..views import styles
import cadnano2.util as util
from .. import cadnano
import os.path, zipfile, shutil, platform, subprocess, tempfile, errno
util.qtWrapImport('QtCore', globals(), ['QObject', 'QSettings', 'pyqtSlot', 'Qt'])
util.qtWrapImport('QtWidgets', globals(), ['QWidget', 'QDialogButtonBox',
                                           'QTableWidgetItem', 'QFileDialog',
                                           'QMessageBox'])


class Preferences(object):
    """docstring for Preferences"""
    def __init__(self):
        self.qs = QSettings()
        self.uiPrefs = Ui_Preferences()
        self.widget = QWidget()
        self.uiPrefs.setupUi(self.widget)
        self.readPreferences()
        self.widget.addAction(self.uiPrefs.actionClose)
        self.uiPrefs.actionClose.triggered.connect(self.hideDialog)
        self.uiPrefs.honeycombRowsSpinBox.valueChanged.connect(self.setHoneycombRows)
        self.uiPrefs.honeycombColsSpinBox.valueChanged.connect(self.setHoneycombCols)
        self.uiPrefs.honeycombStepsSpinBox.valueChanged.connect(self.setHoneycombSteps)
        self.uiPrefs.squareRowsSpinBox.valueChanged.connect(self.setSquareRows)
        self.uiPrefs.squareColsSpinBox.valueChanged.connect(self.setSquareCols)
        self.uiPrefs.squareStepsSpinBox.valueChanged.connect(self.setSquareSteps)
        self.uiPrefs.autoScafComboBox.currentIndexChanged.connect(self.setAutoScaf)
        self.uiPrefs.defaultToolComboBox.currentIndexChanged.connect(self.setStartupTool)
        self.uiPrefs.zoomSpeedSlider.valueChanged.connect(self.setZoomSpeed)
        # self.uiPrefs.helixAddCheckBox.toggled.connect(self.setZoomToFitOnHelixAddition)
        self.uiPrefs.buttonBox.clicked.connect(self.handleButtonClick)
        self.uiPrefs.addPluginButton.clicked.connect(self.addPlugin)

    def showDialog(self):
        # self.exec_()
        self.readPreferences()
        self.widget.show()  # launch prefs in mode-less dialog

    def hideDialog(self):
        self.widget.hide()

    # @pyqtSlot(object)
    def handleButtonClick(self, button):
        """
        Restores defaults. Other buttons are ignored because connections
        are already set up in qt designer.
        """
        if self.uiPrefs.buttonBox.buttonRole(button) == QDialogButtonBox.ButtonRole.ResetRole:
            self.restoreDefaults()

    def readPreferences(self):
        self.qs.beginGroup("Preferences")
        self.honeycombRows = self.qs.value("honeycombRows", styles.HONEYCOMB_PART_MAXROWS)
        self.honeycombCols = self.qs.value("honeycombCols", styles.HONEYCOMB_PART_MAXCOLS)
        self.honeycombSteps = self.qs.value("honeycombSteps", styles.HONEYCOMB_PART_MAXSTEPS)
        self.squareRows = self.qs.value("squareRows", styles.SQUARE_PART_MAXROWS)
        self.squareCols = self.qs.value("squareCols", styles.SQUARE_PART_MAXCOLS)
        self.squareSteps = self.qs.value("squareSteps", styles.SQUARE_PART_MAXSTEPS)
        self.autoScafIndex = self.qs.value("autoScaf", styles.PREF_AUTOSCAF_INDEX)
        self.startupToolIndex = self.qs.value("startupTool", styles.PREF_STARTUP_TOOL_INDEX)
        self.zoomSpeed = self.qs.value("zoomSpeed", styles.PREF_ZOOM_SPEED)
        self.zoomOnHelixAdd = self.qs.value("zoomOnHelixAdd", styles.PREF_ZOOM_AFTER_HELIX_ADD)
        self.qs.endGroup()
        self.uiPrefs.honeycombRowsSpinBox.setProperty("value", self.honeycombRows)
        self.uiPrefs.honeycombColsSpinBox.setProperty("value", self.honeycombCols)
        self.uiPrefs.honeycombStepsSpinBox.setProperty("value", self.honeycombSteps)
        self.uiPrefs.squareRowsSpinBox.setProperty("value", self.squareRows)
        self.uiPrefs.squareColsSpinBox.setProperty("value", self.squareCols)
        self.uiPrefs.squareStepsSpinBox.setProperty("value", self.squareSteps)
        self.uiPrefs.autoScafComboBox.setCurrentIndex(self.autoScafIndex)
        self.uiPrefs.defaultToolComboBox.setCurrentIndex(self.startupToolIndex)
        self.uiPrefs.zoomSpeedSlider.setProperty("value", self.zoomSpeed)
        ptw = self.uiPrefs.pluginTableWidget
        loadedPluginPaths = list(cadnano.loadedPlugins.keys())
        ptw.setRowCount(len(loadedPluginPaths))
        for i in range(len(loadedPluginPaths)):
            row = QTableWidgetItem(loadedPluginPaths[i])
            row.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable)
            ptw.setItem(i, 0, row)
        # self.uiPrefs.helixAddCheckBox.setChecked(self.zoomOnHelixAdd)

    def restoreDefaults(self):
        self.uiPrefs.honeycombRowsSpinBox.setProperty("value", styles.HONEYCOMB_PART_MAXROWS)
        self.uiPrefs.honeycombColsSpinBox.setProperty("value", styles.HONEYCOMB_PART_MAXCOLS)
        self.uiPrefs.honeycombStepsSpinBox.setProperty("value", styles.HONEYCOMB_PART_MAXSTEPS)
        self.uiPrefs.squareRowsSpinBox.setProperty("value", styles.SQUARE_PART_MAXROWS)
        self.uiPrefs.squareColsSpinBox.setProperty("value", styles.SQUARE_PART_MAXCOLS)
        self.uiPrefs.squareStepsSpinBox.setProperty("value", styles.SQUARE_PART_MAXSTEPS)
        self.uiPrefs.autoScafComboBox.setCurrentIndex(styles.PREF_AUTOSCAF_INDEX)
        self.uiPrefs.defaultToolComboBox.setCurrentIndex(styles.PREF_STARTUP_TOOL_INDEX)
        self.uiPrefs.zoomSpeedSlider.setProperty("value", styles.PREF_ZOOM_SPEED)
        # self.uiPrefs.helixAddCheckBox.setChecked(styles.PREF_ZOOM_AFTER_HELIX_ADD)

    def setHoneycombRows(self, rows):
        self.honeycombRows = rows
        self.qs.beginGroup("Preferences")
        self.qs.setValue("honeycombRows", self.honeycombRows)
        self.qs.endGroup()

    def setHoneycombCols(self, cols):
        self.honeycombCols = cols
        self.qs.beginGroup("Preferences")
        self.qs.setValue("honeycombCols", self.honeycombCols)
        self.qs.endGroup()

    def setHoneycombSteps(self, steps):
        self.honeycombSteps = steps
        self.qs.beginGroup("Preferences")
        self.qs.setValue("honeycombSteps", self.honeycombSteps)
        self.qs.endGroup()

    def setSquareRows(self, rows):
        self.squareRows = rows
        self.qs.beginGroup("Preferences")
        self.qs.setValue("squareRows", self.squareRows)
        self.qs.endGroup()

    def setSquareCols(self, cols):
        self.squareCols = cols
        self.qs.beginGroup("Preferences")
        self.qs.setValue("squareCols", self.squareCols)
        self.qs.endGroup()

    def setSquareSteps(self, steps):
        self.squareSteps = steps
        self.qs.beginGroup("Preferences")
        self.qs.setValue("squareSteps", self.squareSteps)
        self.qs.endGroup()

    def setAutoScaf(self, index):
        self.autoScafIndex = index
        self.qs.beginGroup("Preferences")
        self.qs.setValue("autoScaf", self.autoScafIndex)
        self.qs.endGroup()

    def setStartupTool(self, index):
        self.startupToolIndex = index
        self.qs.beginGroup("Preferences")
        self.qs.setValue("startupTool", self.startupToolIndex)
        self.qs.endGroup()

    def setZoomSpeed(self, speed):
        self.zoomSpeed = speed
        self.qs.beginGroup("Preferences")
        self.qs.setValue("zoomSpeed", self.zoomSpeed)
        self.qs.endGroup()

    # def setZoomToFitOnHelixAddition(self, checked):
    #     self.zoomOnHelixAdd = checked
    #     self.qs.beginGroup("Preferences")
    #     self.qs.setValue("zoomOnHelixAdd", self.zoomOnHelixAdd)
    #     self.qs.endGroup()

    def getAutoScafType(self):
        return ['Mid-seam', 'Raster'][self.autoScafIndex]

    def getStartupToolName(self):
        return ['Select', 'Pencil', 'Paint', 'AddSeq'][self.startupToolIndex]

    def addPlugin(self):
        fdialog = QFileDialog(
                    self.widget,
                    "Install Plugin",
                    cadnano.path(),
                    "Cadnano Plugins (*.cnp)")
        fdialog.setAcceptMode(QFileDialog.AcceptMode.AcceptOpen)
        fdialog.setWindowFlags(Qt.WindowType.Sheet)
        fdialog.setWindowModality(Qt.WindowModality.WindowModal)
        fdialog.filesSelected.connect(self.addPluginAtPath)
        self.fileopendialog = fdialog
        fdialog.open()

    def addPluginAtPath(self, fname):
        self.fileopendialog.close()
        fname = str(fname[0])
        print("Attempting to open plugin %s"%fname)
        try:
            zf = zipfile.ZipFile(fname, 'r')
        except Exception as e:
            self.failWithMsg("Plugin file seems corrupt: %s."%e)
            return
        tdir = tempfile.mkdtemp()
        try:
            for f in zf.namelist():
                if f.endswith('/'):
                    os.makedirs(os.path.join(tdir,f))
            for f in zf.namelist():
                if not f.endswith('/'):
                    zf.extract(f, tdir)
        except Exception as e:
            self.failWithMsg("Extraction of plugin archive failed: %s."%e)
            return
        filesInZip = [(f, os.path.join(tdir, f)) for f in os.listdir(tdir)]
        try:
            self.confirmDestructiveIfNecessary(filesInZip)
            self.removePluginsToBeOverwritten(filesInZip)
            self.movePluginsIntoPluginsFolder(filesInZip)
        except OSError:
            print("Couldn't copy files into plugin directory, attempting\
                   again after boosting privileges.")
            if platform.system() == 'Darwin':
                self.darwinAuthedMvPluginsIntoPluginsFolder(filesInZip)
            elif platform.system() == 'Linux':
                self.linuxAuthedMvPluginsIntoPluginsFolder(filesInZip)
            else:
                print("Can't boost privelages on platform %s"%platform.system())
        loadedAPlugin = cadnano.loadAllPlugins()
        if not loadedAPlugin:
            print("Unable to load anythng from plugin %s"%fname)
        self.readPreferences()
        shutil.rmtree(tdir)

    def darwinAuthedMvPluginsIntoPluginsFolder(self, filesInZip):
        envirn={"DST":cadnano.path()+'/plugins'}
        srcstr = ''
        for i in range(len(filesInZip)):
            fileName, filePath = filesInZip[i]
            srcstr += ' \\"$SRC' + str(i) + '\\"'
            envirn['SRC'+str(i)] = filePath
        proc = subprocess.Popen(['osascript','-e',\
                          'do shell script "cp -fR ' + srcstr +\
                          ' \\"$DST\\"" with administrator privileges'],\
                          env=envirn)
        retval = self.waitForProcExit(proc)
        if retval != 0:
            self.failWithMsg('cp failed with code %i'%retval)

    def linuxAuthedMvPluginsIntoPluginsFolder(self, filesInZip):
        args = ['gksudo', 'cp', '-fR']
        args.extend(filePath for fileName, filePath in filesInZip)
        args.append(cadnano.path()+'/plugins')
        proc = subprocess.Popen(args)
        retval = self.waitForProcExit(proc)
        if retval != 0:
            self.failWithMsg('cp failed with code %i'%retval)

    def confirmDestructiveIfNecessary(self, filesInZip):
        for fileName, filePath in filesInZip:
            target = os.path.join(cadnano.path(), 'plugins', fileName)
            if os.path.isfile(target):
                return self.confirmDestructive()
            elif os.path.isdir(target):
                return self.confirmDestructive()

    def confirmDestructive(self):
        mb = QMessageBox(self.widget)
        mb.setIcon(QMessageBox.Warning)
        mb.setInformativeText("The plugin you are trying to install\
has already been installed. Replace the currently installed one?")
        mb.setStandardButtons(QMessageBox.No | QMessageBox.Yes)
        mb.exec_()
        return mb.clickedButton() == mb.button(QMessageBox.Yes)
        
    def removePluginsToBeOverwritten(self, filesInZip):
        for fileName, filePath in filesInZip:
            target = os.path.join(cadnano.path(), 'plugins', fileName)
            if os.path.isfile(target):
                os.unlink(target)
            elif os.path.isdir(target):
                shutil.rmtree(target)

    def movePluginsIntoPluginsFolder(self, filesInZip):
        for fileName, filePath in filesInZip:
            target = os.path.join(cadnano.path(), 'plugins', fileName)
            shutil.move(filePath, target)

    def waitForProcExit(self, proc):
        procexit = False
        while not procexit:
            try:
                retval = proc.wait()
                procexit = True
            except OSError as e:
                if e.errno != errno.EINTR:
                    raise ose
        return retval

    def failWithMsg(self, str):
        mb = QMessageBox(self.widget)
        mb.setIcon(QMessageBox.Warning)
        mb.setInformativeText(str)
        mb.buttonClicked.connect(self.closeFailDialog)
        self.failMessageBox = mb
        mb.open()

    def closeFailDialog(self, button):
        self.failMessageBox.close()
        del self.failMessageBox


from cadnano2.views import styles
import cadnano2.util as util
util.qtWrapImport('QtCore', globals(), ['QRectF', 'Qt'])
util.qtWrapImport('QtGui', globals(), ['QBrush', 'QFont'])
util.qtWrapImport('QtWidgets', globals(),  ['QColorDialog',
                                            'QGraphicsItem',
                                            'QGraphicsSimpleTextItem'])

_font = QFont(styles.thefont, 12, QFont.Weight.Bold)


class ColorPanel(QGraphicsItem):
    _scafColors = styles.scafColors
    _stapColors = styles.stapColors
    _pen = Qt.PenStyle.NoPen

    def __init__(self, parent=None):
        super(ColorPanel, self).__init__(parent)
        self.rect = QRectF(0, 0, 30, 30)
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIgnoresTransformations)
        self.colordialog = QColorDialog()
        # self.colordialog.setOption(QColorDialog.DontUseNativeDialog)
        self._scafColorIndex = -1  # init on -1, painttool will cycle to 0
        self._stapColorIndex = -1  # init on -1, painttool will cycle to 0
        self._scafColor = self._scafColors[self._scafColorIndex]
        self._stapColor = self._stapColors[self._stapColorIndex]
        self._scafBrush = QBrush(self._scafColor)
        self._stapBrush = QBrush(self._stapColor)
        self._initLabel()
        self.hide()

    def _initLabel(self):
        self._label = label = QGraphicsSimpleTextItem("scaf\nstap", parent=self)
        label.setPos(32, 0)
        label.setFont(_font)
        # label.setBrush(_labelbrush)
        # label.hide()

    def boundingRect(self):
        return self.rect

    def paint(self, painter, option, widget=None):
        painter.setPen(self._pen)
        painter.setBrush(self._scafBrush)
        painter.drawRect(0, 0, 30, 15)
        painter.setBrush(self._stapBrush)
        painter.drawRect(0, 15, 30, 15)

    def nextColor(self):
        self._stapColorIndex += 1
        if self._stapColorIndex == len(self._stapColors):
            self._stapColorIndex = 0
        self._stapColor = self._stapColors[self._stapColorIndex]
        self._stapBrush.setColor(self._stapColor)
        self.update()

    def prevColor(self):
        self._stapColorIndex -= 1

    def color(self):
        return self._stapColor

    def scafColorName(self):
        return self._scafColor.name()

    def stapColorName(self):
        return self._stapColor.name()

    def changeScafColor(self):
        self.update()

    def changeStapColor(self):
        self._stapColor = self.colordialog.currentColor()
        self._stapBrush = QBrush(self._stapColor)
        self.update()

    def mousePressEvent(self, event):
        if event.pos().y() < 10:
            newColor = self.colordialog.getColor(self._scafColor)
            if newColor.isValid() and newColor.name() != self._scafColor.name():
                self._scafColor = newColor
                self._scafBrush = QBrush(newColor)
                if not newColor in self._scafColors:
                    self._scafColors.insert(self._scafColorIndex, newColor)
                self.update()
        else:
            newColor = self.colordialog.getColor(self._stapColor)
            if newColor.isValid() and newColor.name() != self._stapColor.name():
                self._stapColor = newColor
                self._stapBrush = QBrush(newColor)
                if not newColor in self._stapColors:
                    self._stapColors.insert(self._stapColorIndex, newColor)
                self.update()




import cadnano2.util as util
util.qtWrapImport('QtCore', globals(), ['QObject', 'pyqtSignal', 'Qt'])
util.qtWrapImport('QtWidgets', globals(), ['QGraphicsObject'])
util.qtWrapImport('QtSvg', globals(), ['QSvgRenderer'])


class SVGButton(QGraphicsObject):
    def __init__(self, fname, parent=None):
        super(SVGButton, self).__init__(parent)
        self.svg = QSvgRenderer(fname)

    def paint(self, painter, options, widget):
        self.svg.render(painter, self.boundingRect())

    def boundingRect(self):
        return self.svg.viewBoxF()

    clicked = pyqtSignal()
    def mousePressEvent(self, event):
        self.clicked.emit()


# Form implementation generated from reading ui file 'dialogs/about.ui'
#
# Created by: PyQt6 UI code generator 6.1.1
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_About(object):
    def setupUi(self, About):
        About.setObjectName("About")
        About.resize(474, 304)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(About.sizePolicy().hasHeightForWidth())
        About.setSizePolicy(sizePolicy)
        About.setSizeGripEnabled(True)
        self.gridLayout = QtWidgets.QGridLayout(About)
        self.gridLayout.setObjectName("gridLayout")
        self.frame = QtWidgets.QFrame(About)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.frame.sizePolicy().hasHeightForWidth())
        self.frame.setSizePolicy(sizePolicy)
        self.frame.setMinimumSize(QtCore.QSize(440, 270))
        self.frame.setFrameShape(QtWidgets.QFrame.Shape.NoFrame)
        self.frame.setFrameShadow(QtWidgets.QFrame.Shadow.Plain)
        self.frame.setObjectName("frame")
        self.appname = QtWidgets.QLabel(self.frame)
        self.appname.setGeometry(QtCore.QRect(0, 10, 191, 31))
        font = QtGui.QFont()
        font.setPointSize(24)
        self.appname.setFont(font)
        self.appname.setObjectName("appname")
        self.version = QtWidgets.QLabel(self.frame)
        self.version.setGeometry(QtCore.QRect(0, 40, 211, 31))
        self.version.setObjectName("version")
        self.info = QtWidgets.QLabel(self.frame)
        self.info.setGeometry(QtCore.QRect(0, 90, 441, 181))
        self.info.setTextFormat(QtCore.Qt.TextFormat.RichText)
        self.info.setAlignment(QtCore.Qt.AlignmentFlag.AlignLeading|QtCore.Qt.AlignmentFlag.AlignLeft|QtCore.Qt.AlignmentFlag.AlignTop)
        self.info.setWordWrap(True)
        self.info.setOpenExternalLinks(True)
        self.info.setObjectName("info")
        self.gridLayout.addWidget(self.frame, 0, 0, 1, 1)

        self.retranslateUi(About)
        QtCore.QMetaObject.connectSlotsByName(About)

    def retranslateUi(self, About):
        _translate = QtCore.QCoreApplication.translate
        About.setWindowTitle(_translate("About", "About Cadnano2"))
        self.appname.setText(_translate("About", "Cadnano2"))
        self.version.setText(_translate("About", "version 2.4"))
        self.info.setText(_translate("About", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:\'.SF NS Text\'; font-size:13pt; font-weight:400; font-style:normal;\">\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-family:\'Lucida Grande\';\">Copyright © 2009–2021 cadnano.org.</span></p>\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:\'Lucida Grande\';\"><br /></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-family:\'Lucida Grande\';\">cadnano is free and open-source. The source code is available under the MIT license.</span></p>\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:\'Lucida Grande\';\"><br /></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-family:\'Lucida Grande\';\">The development has been supported by Wyss Institute at Harvard University and the Dana-Farber Cancer Institute, and the laboratories of William Shih and George Church.</span></p>\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:\'Lucida Grande\';\"><br /></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-family:\'Lucida Grande\';\">This release is a port to PyQt6 by the DouglasLab at UCSF.</span></p></body></html>"))



# Copyright 2011 Autodesk, Inc.  All rights reserved.
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# http://www.opensource.org/licenses/mit-license.php

"""
helixManip.py
Created by Alex Tessier on 2011-08
A Maya manip for controlling Helix Shape similar to the selectTool
"""

import sys
import maya.OpenMaya as OpenMaya
import maya.OpenMayaUI as OpenMayaUI
import maya.OpenMayaMPx as OpenMayaMPx
import maya.cmds as cmds
from cadnano2.cadnano import app

# for controlling the nodes
from cadnano2.model.enum import StrandType
from controllers.mayacontrollers.mayaObjectManager import Mom


helixManipId = OpenMaya.MTypeId(0x00117702)
contextCmdName = "spHelixManipCtxCmd"
nodeName = "spHelixManip"


def project(a, b):
    c = OpenMaya.MVector()
    bn = b.normal()
    c.x = a.x * bn.x
    c.y = a.y * bn.y
    c.z = a.z * bn.z
    return c


class helixManip(OpenMayaMPx.MPxManipContainer):

    class helix():
        id = None
        helixName = ""
        helixNode = OpenMaya.MObject()
        helixTransform = OpenMaya.MObject()

        def __str__(self):
            return self.helixName

    startBaseFloatAttr = OpenMaya.MObject()
    endBaseFloatAttr = OpenMaya.MObject()

    frontDir = OpenMaya.MVector()
    backDir = OpenMaya.MVector()
    manipHandleOffset = 1
    fDistanceFrontManip = OpenMaya.MDagPath()
    fDistanceBackManip = OpenMaya.MDagPath()
    sp = OpenMaya.MPoint()
    cp = OpenMaya.MPoint()
    ep = OpenMaya.MPoint()
    ecp = OpenMaya.MPoint()

    isMouseDown = False

    deltaFront = 0
    deltaBack = 0
    # min/max range of delta (for different length strands)
    maxDelta = 0
    minDelta = 0

    helices = {}
    helicesNames = []
    ffpIdxMap = {}
    epIdxMap = {}

    frontDistance = 0
    backDistance = 0

    firstHelix = 0

    #static
    helixDragMarker = None
    dragMarkerName = "HelixManipDragMarker"
    transformName = dragMarkerName + "Transform"
    meshName = dragMarkerName + "Mesh"

    def createHelixDragMarker(self):

        if not helixManip.helixDragMarker == None:
            return helixManip.helixDragMarker

        try:
            cmds.createNode(
                        "transform",
                        name=helixManip.transformName,
                        skipSelect=True)
            cmds.setAttr("%s.rotateX" % helixManip.transformName, 90)

            cmds.createNode(
                        "mesh",
                        name=helixManip.meshName,
                        parent=helixManip.transformName,
                        skipSelect=True)
            cmds.createNode(
                        "polyPlane",
                        name=helixManip.dragMarkerName,
                        skipSelect=True)
            #cmds.setAttr("%s.width" % dragMarkerName, 1)
            #cmds.setAttr("%s.height" % dragMarkerName, 1)
            cmds.setAttr(
                    "%s.subdivisionsWidth" % helixManip.dragMarkerName, 1)
            cmds.setAttr(
                    "%s.subdivisionsHeight" % helixManip.dragMarkerName, 1)

            cmds.connectAttr("%s.output" % helixManip.dragMarkerName,
                             "%s.inMesh" % helixManip.meshName)

            helixManip.helixDragMarker = helixManip.transformName

        except:
            sys.stderr.write(
                        "helixManip: Error createHelixDragMarker\n")
            sys.stderr.write(sys.exc_info()[0])
            raise

        return helixManip.helixDragMarker

    def moveHelixDragMarkerTo(self, translation):
        cmds.setAttr(
            helixManip.transformName + ".translate",
            translation[0],
            translation[1],
            translation[2],
            type="double3")

    def __init__(self):
        OpenMayaMPx.MPxManipContainer.__init__(self)

    def addHelix(self, HNumber):

        m = Mom()

        h = self.helix()
        h.id = HNumber
        h.helixTransform = m.getNodeFromName(
                                    "%s%s" % (m.helixTransformName, HNumber))
        h.helixNode = m.getNodeFromName("%s%s" % (m.helixNodeName, HNumber))
        nodeFn = OpenMaya.MFnDependencyNode(h.helixNode)
        h.helixName = nodeFn.name()
        self.helices[h.id] = h

        self.helicesNames.append(h.helixName)

        # we only connect the first node, so that the update events trigger
        # the delta value will be the same for all the helices
        if(len(self.helices) == 1):
            self.firstHelix = h.id
            self.connectToDependNode(h)

    def finishedAddingHelices(self):
        self.finishAddingManips()
        #for (id, helix) in self.helices.iteritems():
        OpenMayaMPx.MPxManipContainer.connectToDependNode(
                                self, self.helices[self.firstHelix].helixNode)

    def createChildren(self):
        self.helices = {}
        self.helicesNames = []

        #print "helixManip: createChildren being called..."
        # startPoint should correspond to the end of the helix
        # read the attribute to get the offset from the starting position
        self.fDistanceFrontManip = self.addDistanceManip(
                                                "distanceManip", "distance")
        distanceManipFn = OpenMayaUI.MFnDistanceManip(self.fDistanceFrontManip)
        startPoint = OpenMaya.MPoint(0.0, 0.0, 0.0)
        self.frontDir = OpenMaya.MVector(0.0, 0.0, 1.0)
        distanceManipFn.setStartPoint(startPoint)
        distanceManipFn.setDirection(self.frontDir)

        self.fDistanceBackManip = self.addDistanceManip(
                                                "distancgeteManip", "distance")
        distanceManipFn = OpenMayaUI.MFnDistanceManip(self.fDistanceBackManip)
        startPoint = OpenMaya.MPoint(0.0, 0.0, 0.0)
        self.backDir = OpenMaya.MVector(0.0, 0.0, -1.0)
        distanceManipFn.setStartPoint(startPoint)
        distanceManipFn.setDirection(self.backDir)

    def draw(self, view, path, style, status):
        OpenMayaMPx.MPxManipContainer.draw(self, view, path, style, status)

        if not self.isMouseDown:
            return

        u = OpenMaya.MPoint()
        v = OpenMaya.MPoint()
        drawText = ""
        am = self.activeManip()
        m = self.getTransformMtxFromNode(
                                self.helices[self.firstHelix].helixTransform)
        if am is self.fDistanceFrontManip:
            drawText = str(self.deltaFront)
            if self.deltaFront > 0:
                drawText = "+" + drawText
            c = self.canMove(self.deltaFront)
            if c[0]:
                drawText = "<  " + drawText
            if c[1]:
                drawText = drawText + "  >"
            u = self.sp * m
            v = u + self.frontDir * self.frontDistance
        elif am is self.fDistanceBackManip:
            drawText = str(self.deltaBack)
            if self.deltaBack > 0:
                drawText = "+" + drawText
            c = self.canMove(self.deltaBack)
            if c[0]:
                drawText = "<  " + drawText
            if c[1]:
                drawText = drawText + "  >"
            u = self.ep * m
            v = u + self.backDir * self.backDistance

        w = OpenMaya.MPoint((u.x + v.x) / 2, (u.y + v.y) / 2, (u.z + v.z) / 2)

        view.beginGL()
        view.setDrawColor(OpenMaya.MColor(0.9, 0, 0))
        view.drawText(drawText, w, OpenMayaUI.M3dView.kCenter)
        view.endGL()

    def connectToDependNode(self, helix):
        #print "connectToDependNode called"
        nodeFn = OpenMaya.MFnDependencyNode(helix.helixNode)

        try:
            frontPlug = nodeFn.findPlug("startBaseFloat")
            backPlug = nodeFn.findPlug("endBaseFloat")

            frontManip = OpenMayaUI.MFnDistanceManip(self.fDistanceFrontManip)
            frontManip.connectToDistancePlug(frontPlug)

            backManip = OpenMayaUI.MFnDistanceManip(self.fDistanceBackManip)
            backManip.connectToDistancePlug(backPlug)

            # hookup the converter functions
            self.addPlugToManipConversion(frontManip.startPointIndex())
            self.addPlugToManipConversion(frontManip.currentPointIndex())
            self.addPlugToManipConversion(backManip.startPointIndex())
            self.addPlugToManipConversion(backManip.currentPointIndex())

            idx = self.addManipToPlugConversion(frontPlug)
            self.ffpIdxMap[idx] = helix
            idx = self.addManipToPlugConversion(backPlug)
            self.epIdxMap[idx] = helix

            # initially, current points and start/end points match
            startPos = nodeFn.findPlug("startPos")
            startPoint = self.getFloat3PlugValue(startPos)
            self.sp = startPoint
            self.cp = startPoint + self.frontDir * self.manipHandleOffset

            endPos = nodeFn.findPlug("endPos")
            endPoint = self.getFloat3PlugValue(endPos)
            self.ep = endPoint
            self.ecp = endPoint + self.backDir * self.manipHandleOffset

        except:
            sys.stderr.write(
                        "helixManip: Error finding and connecting plugs\n")
            sys.stderr.write(sys.exc_info()[0])
            raise

    def computeDistance(self, helix, sp, cp, dirv):
        # sp - start point, cp - current point
        u = OpenMaya.MVector(sp)
        v = OpenMaya.MVector(cp)
        w = u - v
        m = self.getTransformMtxFromNode(helix.helixTransform)
        wm = w * m
        if(wm.normal() * dirv.normal() > 0):  # dotproduct
            dir = -1
        else:
            dir = 1
        distance = w.length() * dir
        return distance

    def activeManip(self):

        try:
            activeManip = OpenMaya.MObject()
            self.isManipActive(OpenMaya.MFn.kDistanceManip, activeManip)

            m = Mom()
            if activeManip == self.fDistanceFrontManip.node():
                return self.fDistanceFrontManip
            elif activeManip == self.fDistanceBackManip.node():
                return self.fDistanceBackManip
        except:
            sys.stderr.write("ERROR: helixManip.activeManip\n")
            raise

        return None

    def calculateDeltaBounds(self):
        try:

            self.minDelta = float('-inf')
            self.maxDelta = float('inf')

            am = self.activeManip()
            for (id, helix) in self.helices.items():
                strand = self.getStrand(helix)
                lowIdx, highIdx = strand.idxs()

                minVal = 0
                maxVal = 0
                if am is self.fDistanceFrontManip:
                    lbound, ubound = strand.getResizeBounds(lowIdx)
                    minVal = lbound - lowIdx
                    maxVal = ubound - lowIdx
                elif am is self.fDistanceBackManip:
                    lbound, ubound = strand.getResizeBounds(highIdx)
                    minVal = lbound - highIdx
                    maxVal = ubound - highIdx
                    
                self.minDelta = max(minVal, self.minDelta)
                self.maxDelta = min(maxVal, self.maxDelta)

        except:
            print("calculateDeltaBounds failed!")
            raise

    def snapToXover(self, delta):

        am = self.activeManip()
        newDelta = delta

        try:
            for (id, helix) in self.helices.items():
                strand = self.getStrand(helix)
                idxL, idxH = strand.idxs()
                # idxL = idxL+delta if value[0] else idxL
                # idxH = idxH+delta if value[1] else idxH
                if am is self.fDistanceFrontManip and strand.connectionLow():
                    part = strand.virtualHelix().part()
                    newDelta = part.xoverSnapTo(strand, idxL, delta) - idxL
                elif am is self.fDistanceBackManip and strand.connectionHigh():
                    part = strand.virtualHelix().part()
                    newDelta = part.xoverSnapTo(strand, idxH, delta) - idxH
        except:
            sys.stderr.write("ERROR: helixManip.snapToXover\n")
            raise

        return newDelta

    def canMove(self, delta):

        dirLeft = False
        dirRight = False

        if self.snapToXover(delta) > self.snapToXover(self.minDelta):
            dirLeft = True
        if self.snapToXover(delta) < self.snapToXover(self.maxDelta):
            dirRight = True

        return (dirLeft, dirRight)

    def doPress(self):

        # print "PRESS"
        self.isMouseDown = True

        am = self.activeManip()

        m = Mom()
        m.strandsSelected(
                        self.helicesNames,
                        (am is self.fDistanceFrontManip,
                        am is self.fDistanceBackManip)
                        )

        self.createHelixDragMarker()
        selectedItems = cmds.ls(Mom().helixTransformName + "*", selection=True)
        bbox = cmds.exactWorldBoundingBox(selectedItems)

        cmds.setAttr(
                helixManip.transformName + ".scale",
                bbox[3] - bbox[0],
                0,
                bbox[4] - bbox[1],
                type="double3")

        z = 0
        if am is self.fDistanceFrontManip:
            z = bbox[5]
        elif am is self.fDistanceBackManip:
            z = bbox[2]

        self.moveHelixDragMarkerTo((
                                (bbox[0] + bbox[3]) / 2,
                                (bbox[1] + bbox[4]) / 2,
                                z
                                ))
        cmds.showHidden(helixManip.transformName)

        self.calculateDeltaBounds()

        return OpenMaya.kUnknownParameter

    def doRelease(self):
        # print "RELEASED"

        self.isMouseDown = False

        if(self.deltaFront != 0):
            app().activeDocument.document().resizeSelection(self.deltaFront)

        if(self.deltaBack != 0):
            app().activeDocument.document().resizeSelection(self.deltaBack)

        m = Mom()
        m.strandsSelected(self.helicesNames, (True, True))

        cmds.hide(helixManip.transformName)

        self.frontDistance = 0
        self.backDistance = 0

        return OpenMaya.kUnknownParameter

    def doDrag(self):
        #print "DRAGGING"

        am = self.activeManip()
        selectedItems = cmds.ls(Mom().helixTransformName + "*", selection=True)
        bbox = cmds.exactWorldBoundingBox(selectedItems)
        z = 0
        if am is self.fDistanceFrontManip:
            self.moveHelixDragMarkerTo((
                                (bbox[0] + bbox[3]) / 2,
                                (bbox[1] + bbox[4]) / 2,
                                bbox[5] + self.frontDistance
                                ))
        elif am is self.fDistanceBackManip:
            self.moveHelixDragMarkerTo((
                                (bbox[0] + bbox[3]) / 2,
                                (bbox[1] + bbox[4]) / 2,
                                bbox[2] - self.backDistance
                                ))

        return OpenMaya.kUnknownParameter

    def plugToManipConversion(self, manipIndex):

        #print "plugToManipCalled"

        manipData = OpenMayaUI.MManipData()
        try:

            frontManip = OpenMayaUI.MFnDistanceManip(self.fDistanceFrontManip)
            backManip = OpenMayaUI.MFnDistanceManip(self.fDistanceBackManip)

            boundBoxCenter = OpenMaya.MVector()
            boundBoxScale = OpenMaya.MVector()

            selectedItems = cmds.ls(
                                Mom().helixTransformName + "*", selection=True)
            bbox = cmds.exactWorldBoundingBox(selectedItems)
            boundBoxCenter.x = float((bbox[0] + bbox[3]) / 2)
            boundBoxCenter.y = float((bbox[1] + bbox[4]) / 2)
            boundBoxCenter.z = float((bbox[2] + bbox[5]) / 2)
            boundBoxScale.x = float(bbox[3] - bbox[0])
            boundBoxScale.y = float(bbox[4] - bbox[1])
            boundBoxScale.z = float(bbox[5] - bbox[2])

            if(manipIndex == frontManip.currentPointIndex()):

                ws = boundBoxCenter + project(boundBoxScale / 2, self.frontDir)
                ws += self.frontDir * \
                                (self.manipHandleOffset + self.frontDistance)

                numData = OpenMaya.MFnNumericData()
                numDataObj = numData.create(OpenMaya.MFnNumericData.k3Double)
                numData.setData3Double(ws.x, ws.y, ws.z)
                manipData = OpenMayaUI.MManipData(numDataObj)

            elif(manipIndex == backManip.currentPointIndex()):

                ws = boundBoxCenter + project(boundBoxScale / 2, self.backDir)
                ws += self.backDir * \
                                (self.manipHandleOffset + self.backDistance)

                numData = OpenMaya.MFnNumericData()
                numDataObj = numData.create(OpenMaya.MFnNumericData.k3Double)
                numData.setData3Double(ws.x, ws.y, ws.z)
                manipData = OpenMayaUI.MManipData(numDataObj)

            elif(manipIndex == frontManip.startPointIndex()):

                ws = boundBoxCenter + project(boundBoxScale / 2, self.frontDir)

                numData = OpenMaya.MFnNumericData()
                numDataObj = numData.create(OpenMaya.MFnNumericData.k3Float)
                numData.setData3Float(ws.x, ws.y, ws.z)
                manipData = OpenMayaUI.MManipData(numDataObj)

            elif(manipIndex == backManip.startPointIndex()):

                ws = boundBoxCenter + project(boundBoxScale / 2, self.backDir)

                numData = OpenMaya.MFnNumericData()
                numDataObj = numData.create(OpenMaya.MFnNumericData.k3Float)
                numData.setData3Float(ws.x, ws.y, ws.z)
                manipData = OpenMayaUI.MManipData(numDataObj)
        except:
            sys.stderr.write("ERROR: helixManip.plugToManipConversion\n")
            raise

        return manipData

    def manipToPlugConversion(self, plugIndex):
        #print "manipToPlugConversion", plugIndex

        try:

            if plugIndex in self.ffpIdxMap:  # front float plug
                helix = self.ffpIdxMap[plugIndex]
                numData = OpenMaya.MFnNumericData()
                numDataObj = numData.create(OpenMaya.MFnNumericData.k3Double)

                sp = OpenMaya.MPoint()
                cp = OpenMaya.MPoint()

                frontManip = OpenMayaUI.MFnDistanceManip(
                                                    self.fDistanceFrontManip)

                self.getConverterManipValue(frontManip.startPointIndex(), sp)
                self.getConverterManipValue(frontManip.currentPointIndex(), cp)

                m2 = self.getTransformMtxFromNode(helix.helixTransform)
                m = m2.inverse()

                self.sp = sp * m
                self.cp = cp * m

                # distance is...
                distance = self.computeDistance(
                                        helix, self.sp, self.cp, self.frontDir
                                                ) - self.manipHandleOffset

                self.frontDistance = distance

                delta = self.distanceToBase(helix, distance)
                self.resizeCNHelixFront(helix, delta)

                numData.setData3Double(cp.x, cp.y, cp.z)
                returnData = OpenMayaUI.MManipData(numDataObj)

            elif plugIndex in self.epIdxMap:
                helix = self.epIdxMap[plugIndex]
                numData = OpenMaya.MFnNumericData()
                numDataObj = numData.create(OpenMaya.MFnNumericData.k3Double)

                ep = OpenMaya.MPoint()
                ecp = OpenMaya.MPoint()

                backManip = OpenMayaUI.MFnDistanceManip(
                                                    self.fDistanceBackManip)

                self.getConverterManipValue(backManip.startPointIndex(), ep)
                self.getConverterManipValue(backManip.currentPointIndex(), ecp)

                m2 = self.getTransformMtxFromNode(helix.helixTransform)
                m = m2.inverse()

                self.ep = ep * m
                self.ecp = ecp * m

                # distance is...
                distance = self.computeDistance(
                                        helix, self.ep, self.ecp, self.backDir
                                                ) - self.manipHandleOffset

                self.backDistance = distance

                delta = self.distanceToBase(helix, distance)
                self.resizeCNHelixBack(helix, delta)

                numData.setData3Double(ecp.x, ecp.y, ecp.z)
                returnData = OpenMayaUI.MManipData(numDataObj)
        except:
            sys.stderr.write("ERROR: helixManip.manipToPlugConversion\n")
            raise

        return returnData

    def getTransformFromNode(self, trnsNode):
        try:
            dagFn = OpenMaya.MFnDagNode(trnsNode)
            path = OpenMaya.MDagPath()
            dagFn.getPath(path)
            transformFn = OpenMaya.MFnTransform(path)
            return transformFn
        except:
            print("getTransformFromNode -- unable to get transform from node")
            print(trnsNode)
            raise

    def getTransformMtxFromNode(self, trnsNode):
        mtx = OpenMaya.MMatrix()
        try:
            transformFn = self.getTransformFromNode(trnsNode)
            mm = transformFn.transformation()
            mtx = mm.asMatrix()
        except:
            print("Unable to get transformation matrix for node:")
            print(trnsNode)

        return mtx

    # Getting values out of references is a little crazy in the 1.0
    # Maya Python API....
    def getFloat3PlugValue(self, plug):
        # Retrieve the value as an MObject
        object = plug.asMObject()

        # TODO type checking that our plug is a float3 plug!
        # Convert the MObject to a float3
        numDataFn = OpenMaya.MFnNumericData(object)

        xParam = OpenMaya.MScriptUtil()
        xParam.createFromDouble(0.0)
        xPtr = xParam.asFloatPtr()

        yParam = OpenMaya.MScriptUtil()
        yParam.createFromDouble(0.0)
        yPtr = yParam.asFloatPtr()

        zParam = OpenMaya.MScriptUtil()
        zParam.createFromDouble(0.0)
        zPtr = zParam.asFloatPtr()

        numDataFn.getData3Float(xPtr, yPtr, zPtr)

        return OpenMaya.MPoint(OpenMaya.MScriptUtil.getFloat(xPtr),
                                OpenMaya.MScriptUtil.getFloat(yPtr),
                                OpenMaya.MScriptUtil.getFloat(zPtr))

    def getStrand(self, helix):
        m = Mom()
        return m.mayaToCn[helix.helixName]

    def getRise(self, helix):
        try:
            n = OpenMaya.MFnDependencyNode(helix.helixNode)
            risePlug = n.findPlug("rise")
            rise = risePlug.asDouble()
            return rise
        except:
            print("Failed to retrieve rise from helix %s" % helix.helixName)

    def distanceToBase(self, helix, distance):
        try:
            ## given a distance from the origin of the strand,
            ## compute the change in bases

            rise = self.getRise(helix)
            delta = int(distance / rise)

            return delta
        except:
            print("distanceToBase failed!")

    def baseToDistance(self, helix, base):
        try:
            ## given a distance from the origin of the strand,
            ## compute the change in bases

            rise = self.getRise(helix)
            distance = base * rise

            return distance
        except:
            print("baseToDistance failed!")

    def resizeCNHelixFront(self, helix, delta):
        #print "resizeCNHelixFront"
        try:
            strand = self.getStrand(helix)
            lowIdx, highIdx = strand.idxs()
            newLow = lowIdx - delta
            newLow = min(newLow, highIdx - 1)
            newLow = max(newLow, strand.part().minBaseIdx())

            self.deltaFront = newLow - lowIdx
            self.deltaFront = max(self.minDelta, self.deltaFront)
            self.deltaFront = min(self.maxDelta, self.deltaFront)

            self.deltaFront = self.snapToXover(self.deltaFront)

            self.frontDistance = -self.baseToDistance(helix, self.deltaFront)
        except:
            print("resizeCNHelixFront failed!")
            raise

    def resizeCNHelixBack(self, helix, delta):
        #print "resizeCNHelixBack"
        try:
            strand = self.getStrand(helix)
            lowIdx, highIdx = strand.idxs()
            newHigh = highIdx + delta
            newHigh = min(newHigh, strand.part().maxBaseIdx())
            newHigh = max(newHigh, lowIdx + 1)

            self.deltaBack = newHigh - highIdx
            self.deltaBack = max(self.minDelta, self.deltaBack)
            self.deltaBack = min(self.maxDelta, self.deltaBack)

            self.deltaBack = self.snapToXover(self.deltaBack)

            self.backDistance = self.baseToDistance(helix, self.deltaBack)
        except:
            print("resizeCNHelixBack failed!")
            raise


def helixManipCreator():
    #print "helixManip: helixManipCreator called"
    return OpenMayaMPx.asMPxPtr(helixManip())


def helixManipInitialize():
        #print "helixManip: helixManipInitialize called"
        OpenMayaMPx.MPxManipContainer.initialize()
        typedAttr = OpenMaya.MFnTypedAttribute()
        nAttr = OpenMaya.MFnNumericAttribute()

        helixManip.startBaseFloatAttr = \
            nAttr.create('helixManipStartPos',
            'hmsp',
            OpenMaya.MFnNumericData.kFloat,
            0.0)

        nAttr.setWritable(True)
        nAttr.setStorable(False)
        helixManip.endBaseFloatAttr = \
            nAttr.create('helixManipEndPos',
            'hmep',
            OpenMaya.MFnNumericData.kFloat,
            0.0)

        nAttr.setWritable(True)
        nAttr.setStorable(False)

        helixManip.addAttribute(helixManip.startBaseFloatAttr)
        helixManip.addAttribute(helixManip.endBaseFloatAttr)


# initialize the script plug-in
def initializePlugin(mobject):
    #print "helixManip: initializePlugin for helixManip called"
    mplugin = OpenMayaMPx.MFnPlugin(mobject)
    try:
        mplugin.registerNode(nodeName, helixManipId, helixManipCreator,
                    helixManipInitialize, OpenMayaMPx.MPxNode.kManipContainer)
    except:
        print("helixManip: Failed to register node: %s" % nodeName)
        raise


# uninitialize the script plug-in
def uninitializePlugin(mobject):
    #print "helixManip uninitializePlugin for helixManip called"
    mplugin = OpenMayaMPx.MFnPlugin(mobject)

    try:
        mplugin.deregisterNode(helixManipId)
    except:
        print("helixManip: Failed to deregister node: %s" % nodeName)
        raise



#!/usr/bin/env python
# encoding: utf-8


class Insertion(object):
    """
    Insertions do affect an applied sequence and do not store a sequence
    themselves.  They are a skip if the length is less than 0
    """
    def __init__(self, index, length):
        self._length = length
        self._index  = index
    # end def

    def length(self):
        """
        This is the length of a sequence that is immutable by the strand
        """
        return self._length

    def setLength(self, length):
        self._length = length
    # end def

    def updateIdx(self, delta):
        self._index += delta
    # end def

    def idx(self):
        return self._index
    # end def

    def isSkip(self):
        return self.length() < 0
# end class


    """
functionaltests.py

Created by Shawn Douglas on 2011-06-28.
"""

import sys
sys.path.insert(0, '.')

import time
from PyQt4.QtCore import Qt, QPoint
from cadnano2.data.dnasequences import sequences
from cadnano2.model.enum import StrandType
from cadnano2.model.virtualhelix import VirtualHelix
from tests.cadnanoguitestcase import CadnanoGuiTestCase
import tests.cadnanoguitestcase  # for main()


class FunctionalTests(CadnanoGuiTestCase):
    """
    Functional tests are end-to-end tests that simulate user interaction
    with the interface and verify that the final outputs (e.g. staple
    sequences) are correct.

    Run these tests by calling "python -m tests.functionaltests" from cadnano2
    root directory.
    """
    def setUp(self):
        """
        The setUp method is called before running any test. It is used
        to set the general conditions for the tests to run correctly.
        """
        CadnanoGuiTestCase.setUp(self)
        # Add extra initialization here

    def tearDown(self):
        """
        The tearDown method is called at the end of running each test,
        generally used to clean up any objects created in setUp
        """
        CadnanoGuiTestCase.tearDown(self)
        # Add functional-test-specific cleanup here

    def testFunctional1(self):
        """docstring for testFunctional1"""
        pass

    def getTestSequences(self, designname, sequencesToApply):
        """
        Called by a sequence-verification functional test to read in a file
        (designname), apply scaffold sequence(s) to that design, and return
        the set of staple sequences."""
        # set up the document
        from cadnano2.model.io.decoder import decode

        inputfile = "tests/functionaltestinputs/%s" % designname
        document = self.documentController.document()
        with file(inputfile) as f:
            decode(document, f.read())
        self.setWidget(self.documentController.win, False, None)
        part = document.selectedPart()
        # apply one or more sequences to the design
        for sequenceName, startVhNum, startIdx in sequencesToApply:
            sequence = sequences.get(sequenceName, None)
            for vh in part.getVirtualHelices():
                if vh.number() == startVhNum:
                    strand = vh.scaffoldStrandSet().getStrand(startIdx)
                    strand.oligo().applySequence(sequence)
        generatedSequences = part.getStapleSequences()
        return set(generatedSequences.splitlines())

    def getRefSequences(self, designname):
        """docstring for getRefSequences"""
        staplefile = "tests/functionaltestinputs/%s" % designname
        with open(staplefile, 'rU') as f:
            readSequences = f.read()
        return set(readSequences.splitlines())

    ####################### Staple Comparison Tests ########################
    def testStapleOutput_simple42legacy(self):
        """p7308 applied to 42-base duplex (json source)"""
        designname = "simple42legacy.json"
        refname = "simple42legacy.csv"
        sequences = [("p7308", 0, 0)]
        testSet = self.getTestSequences(designname, sequences)
        refSet = self.getRefSequences(refname)
        self.assertEqual(testSet, refSet)

    def testStapleOutput_insert_size_1(self):
        """Test sequence output with a single insert of size 1"""
        designname = "loop_size_1.json"
        refname = "loop_size_1.csv"
        sequences = [("M13mp18", 0, 14)]
        testSet = self.getTestSequences(designname, sequences)
        refSet = self.getRefSequences(refname)
        self.assertEqual(testSet, refSet)

    def testStapleOutput_skip(self):
        """Simple design with a single skip"""
        designname = "skip.json"
        refname = "skip.csv"
        sequences = [("M13mp18", 0, 14)]
        testSet = self.getTestSequences(designname, sequences)
        refSet = self.getRefSequences(refname)
        self.assertEqual(testSet, refSet)

    def testStapleOutput_inserts_and_skips(self):
        """Insert and skip stress test"""
        designname = "loops_and_skips.json"
        refname = "loops_and_skips.csv"
        sequences = [("M13mp18", 0, 0)]
        testSet = self.getTestSequences(designname, sequences)
        refSet = self.getRefSequences(refname)
        self.assertEqual(testSet, refSet)

    def testStapleOutput_Nature09_monolith(self):
        """Staples match reference set for Nature09 monolith"""
        designname = "Nature09_monolith.json"
        refname = "Nature09_monolith.csv"
        sequences = [("p7560", 4, 73)]
        testSet = self.getTestSequences(designname, sequences)
        refSet = self.getRefSequences(refname)
        self.assertEqual(testSet, refSet)

    def testStapleOutput_Nature09_squarenut(self):
        """Staples match reference set for Nature09 squarenut"""
        designname = "Nature09_squarenut.json"
        refname = "Nature09_squarenut.csv"
        sequences = [("p7560", 15, 100)]
        testSet = self.getTestSequences(designname, sequences)
        refSet = self.getRefSequences(refname)
        self.assertEqual(testSet, refSet)

    def testStapleOutput_Science09_prot120_98_v3(self):
        """Staples match reference set for Science09 protractor 120 v3"""
        designname = "Science09_prot120_98_v3.json"
        refname = "Science09_prot120_98_v3.csv"
        sequences = [("p7704", 0, 105)]
        testSet = self.getTestSequences(designname, sequences)
        refSet = self.getRefSequences(refname)
        self.assertEqual(testSet, refSet)

    def testStapleOutput_Science09_beachball_v1_json(self):
        """Staples match reference set for Science09 beachball (json source)"""
        designname = "Science09_beachball_v1.json"
        refname = "Science09_beachball_v1.csv"
        sequences = [("p7308", 10, 221)]
        testSet = self.getTestSequences(designname, sequences)
        refSet = self.getRefSequences(refname)
        self.assertEqual(testSet, refSet)

    def testStapleOutput_Gap_Vs_Skip(self):
        """Staple gap output as '?'; staple skip output as ''"""
        designname = "gap_vs_skip.json"
        refname = "gap_vs_skip.csv"
        sequences = [("M13mp18", 0, 11), ("M13mp18", 2, 11)]
        testSet = self.getTestSequences(designname, sequences)
        refSet = self.getRefSequences(refname)
        self.assertEqual(testSet, refSet)

    ####################### Standard Functional Tests ########################
    # def testActiveSliceHandleAltShiftClick(self):
    #     """Alt+Shift+Click on ActiveSliceHandle extends scaffold strands."""
    #     # Create a new Honeycomb part
    #     newHoneycombPartButton = self.mainWindow.topToolBar.widgetForAction(\
    #                                    self.mainWindow.actionNewHoneycombPart)
    #     self.click(newHoneycombPartButton)
    #     # Click each SliceHelix
    #     sliceGraphicsItem = self.documentController.sliceGraphicsItem
    #     slicehelix1 = sliceGraphicsItem.getSliceHelixByCoord(0, 0)
    #     slicehelix2 = sliceGraphicsItem.getSliceHelixByCoord(0, 1)
    #     self.click(slicehelix1, qgraphicsscene=self.mainWindow.slicescene)
    #     self.click(slicehelix2, qgraphicsscene=self.mainWindow.slicescene)
    #     # Click the activeSliceHandle with ALT and SHIFT modifiers
    #     pathHelixGroup = self.documentController.pathHelixGroup
    #     activeSliceHandle = pathHelixGroup.activeSliceHandle()
    #     self.mousePress(activeSliceHandle,\
    #                     modifiers=Qt.KeyboardModifier.AltModifier|Qt.KeyboardModifier.ShiftModifier,\
    #                     qgraphicsscene=self.mainWindow.pathscene)
    #     # Check the model for correctness
    #     vh0 = self.app.v[0]
    #     vh1 = self.app.v[1]
    #     str0 = "0 Scaffold: _,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,_\n0 Staple:   _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_"
    #     str1 = "1 Scaffold: _,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,_\n1 Staple:   _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_"
    #     self.assertEqual(repr(vh0), str0)
    #     self.assertEqual(repr(vh1), str1)
    # 
    # def testEndpointAltClick(self):
    #     """Alt+Click on a 5' or 3' endpoint extends it to its boundary."""
    #     # Create a new Honeycomb part
    #     newHoneycombPartButton = self.mainWindow.topToolBar.widgetForAction(\
    #                                    self.mainWindow.actionNewHoneycombPart)
    #     self.click(newHoneycombPartButton)
    #     # Click each SliceHelix
    #     sliceGraphicsItem = self.documentController.sliceGraphicsItem
    #     slicehelix1 = sliceGraphicsItem.getSliceHelixByCoord(0, 0)
    #     slicehelix2 = sliceGraphicsItem.getSliceHelixByCoord(0, 1)
    #     self.mousePress(slicehelix1, qgraphicsscene=self.mainWindow.slicescene)
    #     self.mousePress(slicehelix2, qgraphicsscene=self.mainWindow.slicescene)
    #     # Click the path helices with the ALT modifier
    #     pathHelixGroup = self.documentController.pathHelixGroup
    #     ph0 = pathHelixGroup.getPathHelix(0)
    #     ph1 = pathHelixGroup.getPathHelix(1)
    #     self.mousePress(ph0, position=QPoint(410, 10),\
    #                     modifiers=Qt.KeyboardModifier.AltModifier,\
    #                     qgraphicsscene=self.mainWindow.pathscene)
    #     self.mousePress(ph0, position=QPoint(450, 10),\
    #                     modifiers=Qt.KeyboardModifier.AltModifier,\
    #                     qgraphicsscene=self.mainWindow.pathscene)
    #     self.mousePress(ph1, position=QPoint(410, 30),\
    #                     modifiers=Qt.KeyboardModifier.AltModifier,\
    #                     qgraphicsscene=self.mainWindow.pathscene)
    #     self.mousePress(ph1, position=QPoint(450, 30),\
    #                     modifiers=Qt.KeyboardModifier.AltModifier,\
    #                     qgraphicsscene=self.mainWindow.pathscene)
    #     # Check the model for correctness
    #     vh0 = self.app.v[0]
    #     vh1 = self.app.v[1]
    #     str0 = "0 Scaffold: _,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,_\n0 Staple:   _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_"
    #     str1 = "1 Scaffold: _,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,> <,_\n1 Staple:   _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_ _,_"
    #     self.assertEqual(repr(vh0), str0)
    #     self.assertEqual(repr(vh1), str1)
    #     # self.debugHere()  # Stop simulation and give control to user


if __name__ == '__main__':
    print("Running Functional Tests")
    tests.cadnanoguitestcase.main()



# Copyright 2011 Autodesk, Inc.  All rights reserved.
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# http://www.opensource.org/licenses/mit-license.php

"""
virtualhelixitem.py
Created by Simon Breslav on 2011-08-29.
"""

from string import *
import math
import random
import os
import sys
from cadnano2.views import styles
from cadnano2.model.enum import LatticeType
from cadnano2.model.enum import StrandType
from cadnano2.model.virtualhelix import VirtualHelix
#from cadnano2.model..strands.normalstrand import NormalStrand
#from cadnano2.model..strands.xoverstrand import XoverStrand3, XoverStrand5

from controllers.mayacontrollers.mayaObjectManager import Mom
from cadnano2.controllers.itemcontrollers.virtualhelixitemcontroller \
                                            import VirtualHelixItemController

import maya.OpenMayaUI as mui
import maya.OpenMaya as mo
import maya.cmds as cmds
import cadnano2.util as util

from .stranditem import StrandItem

# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtCore', globals(), ['pyqtSignal', 'pyqtSlot', 'QObject'])
util.qtWrapImport('QtGui', globals(), ['QColor'])


class VirtualHelixItem(object):
    """
    VirtualHelixItem is the container for StrandItems for is the Maya 3D View,
    it does not get visualized in any way right now.
    """
    baseWidth = styles.PATH_BASE_WIDTH

    def __init__(self, partItem, modelVirtualHelix, x, y):
        """
        Initialize member variables.
        Setup VirtualHelixItemController, that takes care of all the
        slots and signals between VirtualHelix model and this VirtualHelixItem.
        """
        self._partItem = partItem
        self._modelVirtualHelix = modelVirtualHelix
        self._x = x
        self._y = y
        coords = modelVirtualHelix.coord()
        self._row = coords[0]
        self._col = coords[1]
        self.strandIDs = []
        self._modState = False
        self._strandItems = {}
        self.stapleIndicatorCount = 0
        self.stapleModIndicatorIDs = []

        self._controller = VirtualHelixItemController(self, modelVirtualHelix)
    # end def

    def partItem(self):
        """PartItem Accessor - parent Item"""
        return self._partItem

    def virtualHelix(self):
        """VirtualHelix Model Accessor"""
        return self._modelVirtualHelix

    def number(self):
        """VirtualHelix Model Index Number Accessor"""
        return self._modelVirtualHelix.number()

    def row(self):
        """Row Accessor"""
        return self.coord()[0]

    def col(self):
        """Col Accessor"""
        return self.coord()[1]

    def x(self):
        """Actual x position in the the Maya 3D View"""
        return self._x
    # end def

    def y(self):
        """Actual y position in the the Maya 3D View"""
        return self._y
    # end def

    def coord(self):
        """Returns a tuple (row, column) of the VitualHelix model"""
        return self._modelVirtualHelix.coord()
    # end def

    def isEvenParity(self):
        """Returns parity of the VitualHelix model"""
        return self._modelVirtualHelix.isEvenParity()

    def StrandIDs(self):
        """
        Returns a list of Strand IDs within this VH, the IDs are suffixes
        of Maya Nodes, and are used to for Maya<->Cadnano communication
        in mayaObjectManager
        """
        return self.strandIDs

    def setModifyState(self, val):
        """Update Modify state for this VirtualHelixItem"""
        self._modState = val

    ### SLOTS ###
    def strandAddedSlot(self, sender, strand):
        """
        Instantiates a StrandItem upon notification that the model has a
        new Strand.  The StrandItem is responsible for creating its own
        controller for communication with the model, and for adding itself to
        its parent (which is *this* VirtualHelixItem, i.e. 'self').
        """
        #print "solidview.VirtualHelixItem.strandAddedSlot"
        m = Mom()
        mID = m.strandMayaID(strand)
        self.strandIDs.append(mID)
        sI = StrandItem(mID, strand, self)
        self._strandItems[sI] = True
        self.updateDecorators()
        #print "solidview.VirtualHelixItem.strandAddedSlot done %s" % mID
    # end def

    @pyqtSlot(object)
    def decoratorAddedSlot(self, decorator):
        """decoratorAddedSlot - empty"""
        pass

    @pyqtSlot(object)
    def virtualHelixNumberChangedSlot(self, virtualHelix):
        """virtualHelixNumberChangedSlot - empty"""
        pass
    # end def

    @pyqtSlot(object)
    def virtualHelixRemovedSlot(self, virtualHelix):
        """
        Clears out private variables and disconnects signals
        """
        #print "solidview.VirtualHelixItem.virtualHelixRemovedSlot"
        self._partItem.removeVirtualHelixItem(self)
        self._partItem = None
        self._modelVirtualHelix = None
        self._controller.disconnectSignals()
        self._controller = None
    # end def

    def updateDecorators(self):
        """
        Clears out Pre-Decorators, and re-populates them if they should be
        visible
        """
        self.clearDecorators()
        if self._modState:
            m = Mom()
            for mID in self.strandIDs:
                mayaNodeInfo = "%s%s" % (m.helixMeshName, mID)
                #print "mayaNodeInfo: %s" % mayaNodeInfo
                strand = m.mayaToCn[mayaNodeInfo]
                if(strand.strandSet().isStaple()):
                    self.createDecorators(strand)

    def cadnanoVBaseToMayaCoords(self, base, strand):
        """
        Given a Strand and a Base, returns a 3D location of that base
        """
        m = Mom()
        mID = m.strandMayaID(strand)
        cylinderName = "%s%s" % (m.helixNodeName, mID)
        if cmds.objExists(cylinderName):
            rise = cmds.getAttr("%s.rise" % cylinderName)
            startBase = cmds.getAttr("%s.startBase" % cylinderName)
            startPos = cmds.getAttr("%s.startPos" % cylinderName)
            base0Pos = startPos[0][1] + (startBase * rise)
            ourPos = base0Pos - (base * rise)
            zComp = ourPos

            rotation = cmds.getAttr("%s.rotation" % cylinderName)
            radius = cmds.getAttr("%s.radius" % cylinderName)
            parity = cmds.getAttr("%s.parity" % cylinderName)
            strandType = cmds.getAttr("%s.strandType" % cylinderName)
            rotationOffset = cmds.getAttr("%s.rotationOffset" % cylinderName)
            decoratorRotOffset = cmds.getAttr("%s.decoratorRotOffset"
                                              % cylinderName)
            # not clear why decoratorRotOffset is not in radians but
            # rotationOffset is
            decoratorRotOffset = decoratorRotOffset * math.pi / 180
            starting_rotation = (math.pi * (not parity)) + rotationOffset + \
                                decoratorRotOffset + \
                                (math.pi * strandType)
            fullrotation = -rotation * base * math.pi / 180
            #print full rotation
            xComp = self._x + radius * \
                    math.cos(starting_rotation + fullrotation)
            yComp = self._y + radius * \
                    math.sin(starting_rotation + fullrotation)
            #print "%f %f %f" % (xComp, yComp, zComp)
            return (xComp, yComp, zComp)
        else:
            raise IndexError

    def clearDecorators(self):
        """Remove all the Pre-Decortators"""
        m = Mom()
        for mID in self.stapleModIndicatorIDs:
            transformName = "%s%s" % (m.decoratorTransformName, mID)
            #print "delete %s" % transformName
            m = Mom()
            m.removeDecoratorMapping(mID)
            if cmds.objExists(transformName):
                cmds.delete(transformName)
        self.stapleModIndicatorIDs = []
        self.stapleIndicatorCount = 0

    def createDecorators(self, strand):
        """Create a set of new Pre-Decortators for a given strand"""
        m = Mom()
        strandId = m.strandMayaID(strand)
        totalNumBases = self._modelVirtualHelix.part().maxBaseIdx()
        preDecoratorIdxList = strand.getPreDecoratorIdxList()

        for baseIdx in preDecoratorIdxList:
            # XXX [SB+AT] NOT THREAD SAFE
            while cmds.objExists("%s%s_%s" % (m.decoratorNodeName,
                                              strandId,
                                              self.stapleIndicatorCount)):
                self.stapleIndicatorCount += 1
            stapleId = "%s_%s" % (strandId, self.stapleIndicatorCount)
            coords = self.cadnanoVBaseToMayaCoords(baseIdx, strand)
            stapleModNodeInfo = self.createDecoratorNodes(coords, stapleId)
            self.stapleModIndicatorIDs.append(stapleId)
            m = Mom()
            m.decoratorToVirtualHelixItem[stapleModNodeInfo[2]] = (self,
                                                                   baseIdx,
                                                                   strand)
            m.decoratorToVirtualHelixItem[stapleModNodeInfo[1]] = (self,
                                                                   baseIdx,
                                                                   strand)

    def createDecoratorNodes(self, coords, mID):
        """Create a actual Maya Nodes for a new Pre-Decortators"""
        m = Mom()
        stapleModIndicatorName = "%s%s" % (m.decoratorNodeName, mID)
        transformName = "%s%s" % (m.decoratorTransformName, mID)
        meshName = "%s%s" % (m.decoratorMeshName, mID)
        shaderName = "%s" % m.decoratorShaderName

        cmds.createNode("transform", name=transformName, skipSelect=True)
        cmds.setAttr("%s.rotateX" % transformName, 90)
        cmds.setAttr("%s.translateX" % transformName, coords[0])
        cmds.setAttr("%s.translateY" % transformName, coords[1])
        cmds.setAttr("%s.translateZ" % transformName, coords[2])
        cmds.createNode("mesh",
                        name=meshName,
                        parent=transformName,
                        skipSelect=True)
        #cmds.createNode("spPreDecoratorNode", name=stapleModIndicatorName)
        cmds.createNode("polySphere",
                        name=stapleModIndicatorName,
                        skipSelect=True)
        cmds.setAttr("%s.radius" % stapleModIndicatorName, .25)
        cmds.setAttr("%s.subdivisionsAxis" % stapleModIndicatorName, 4)
        cmds.setAttr("%s.subdivisionsHeight" % stapleModIndicatorName, 4)

        #cmds.connectAttr("%s.outputMesh" % stapleModIndicatorName,
        #                 "%s.inMesh" % meshName)
        cmds.connectAttr("%s.output" % stapleModIndicatorName,
                         "%s.inMesh" % meshName)

        if not cmds.objExists(shaderName):
            # Shader does not exist create one
            cmds.shadingNode('lambert', asShader=True, name=shaderName)
            cmds.sets(n="%sSG" % shaderName, r=True, nss=True, em=True)
            cmds.connectAttr("%s.outColor" % shaderName,
                             "%sSG.surfaceShader" % shaderName)
            cmds.setAttr("%s.color" % shaderName,
                         0.0, 0.0, 0.0,
                         type="double3")
            cmds.sets(meshName, forceElement="%sSG" % shaderName)
        else:
            #shader exist connect
            cmds.sets(meshName, forceElement="%sSG" % shaderName)
        return (stapleModIndicatorName, transformName, meshName, shaderName)
    # end def

    def removeStrandItem(self, strandItem):
        """Remove a StrandItem from the local list of StrandItems"""
        del self._strandItems[strandItem]
    # end def



###############################################################################
#
# Copyright 2011 Autodesk, Inc.  All rights reserved.
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
###############################################################################


import maya.cmds as cmds
import string
commandWhiteList = [
                "NameComPop_hotBox",

                "NameComUndo",
                "NameComRedo",

                "NameComFit_Selected_in_Active_Panel",
                "NameComFit_Select_in_All_Panels",

                "SelectToolOptionsNameCommand",
                "TranslateToolWithSnapMarkingMenuNameCommand",
                "RotateToolWithSnapMarkingMenuNameCommand",
                "ScaleToolWithSnapMarkingMenuNameCommand",

                "NameComIncrease_Manipulator_Size",
                "NameComDecrease_Manipulator_Size",

                "NameComWireframe_Display",
                "NameComShaded_Display"
                ]
disabledHotKeys = []


class HotKey():
    key = ""
    alt = ""
    ctl = ""
    cmd = ""
    name = None
    rname = None

    def __init__(self, key, name, rname, alt=False, ctl=False, cmd=False):
        self.key = key
        self.name = name
        self.rname = rname
        self.alt = alt
        self.ctl = ctl
        self.cmd = cmd


def saveHotKey(key):
    global disabledHotKeys
    for alt in range(0, 2):
        for ctl in range(0, 2):
            for cmd in range(0, 2):
                name = cmds.hotkey(
                                key, query=True, name=True,
                                alt=alt, ctl=ctl, cmd=cmd)
                rname = cmds.hotkey(
                                key, query=True, releaseName=True,
                                alt=alt, ctl=ctl, cmd=cmd)
                if (name != None or rname != None):
                    disabledHotKeys.append(
                                HotKey(key, name, rname, alt, ctl, cmd))


def disableHotKey(key):
    #print key
    saveHotKey(key)
    for alt in range(0, 2):
        for ctl in range(0, 2):
            for cmd in range(0, 2):
                name = cmds.hotkey(
                                key, q=True, n=True,
                                alt=alt, ctl=ctl, cmd=cmd)
                rname = cmds.hotkey(
                                key, q=True, rn=True,
                                alt=alt, ctl=ctl, cmd=cmd)
                if (name != None or rname != None):
                    if not name in commandWhiteList:
                        cmds.hotkey(k=key, n=None, rn=None,
                                    alt=alt, ctl=ctl, cmd=cmd)

                        #print "disabling alt=%s ctl=%s cmd=%s %s - %s" % \
                        #                    (alt, ctl, cmd, name, rname)
                    else:
                        #print "not disabling %s" % name
                        pass


def disableAllHotKeys():
    #Forces Maya to save prefs
    #Sometimes Maya doesn't save the hotkeys properly
    #   and restarts with them disabled.
    cmds.savePrefs(hotkeys=True)

    """
    This method loops through all the hotkeys we want to disable and calls
    disableHotKey on each of them.

    First we build up our own char list...
    Why not use string.printable? Because it includes unwanted whitespace.
    Why not just use string.letters + string.digits + string.punctuation?
    Because Python in Maya barfs with the following in the lowercase portion
    of string.letters: TypeError: bad argument type for built-in operation.
    """
    letters = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
    digits = '0123456789'
    punctuation = """!"#$%&'()*+,-./:;<=>?@[\]^_`{|}~"""
    chars = digits + punctuation + letters
    for c in chars:
        disableHotKey(c)

    # Function keys F1 through F12
    for key in range(1, 13):
        disableHotKey("F" + str(key))

    # Other named keys
    otherHotKeys = ["Up", "Down", "Right", "Left", "Home", "End",
                    "Page_Up", "Page_Down", "Insert", "Return", "Space"]
    for key in otherHotKeys:
        disableHotKey(key)


def restoreAllHotKeys():
    global disabledHotKeys
    for hotkey in disabledHotKeys:
        cmds.hotkey(
                k=hotkey.key,
                n=hotkey.name,
                rn=hotkey.rname,
                alt=hotkey.alt,
                ctl=hotkey.ctl,
                cmd=hotkey.cmd)
    disabledHotKeys = []

    cmds.savePrefs(hotkeys=True)


from .abstractpathtool import AbstractPathTool
import cadnano2.util as util
util.qtWrapImport('QtCore', globals(), [])
util.qtWrapImport('QtGui', globals(), [])


class SelectTool(AbstractPathTool):
    """
    SelectTool is the default tool. It allows editing of breakpoints
    (by clicking and dragging) and toggling of crossovers.
    """
    def __init__(self, controller):
        super(SelectTool, self).__init__(controller)

    def __repr__(self):
        return "selectTool"  # first letter should be lowercase


#!/usr/bin/env python
# encoding: utf-8


class Modifier(object):
    """
    Modifiers do affect an applied sequence and do not store a sequence
    themselves.  They cause a base changed to another sequence.
    Modifiers DO NOT affect the length of a strand
    """
    def __init__(self, idx):
        if self.__class__ == Modifier:
            e = "Modifier should be subclassed."
            raise NotImplementedError(e)
        self._mType = None
        self._lowIdx  = idx
        self._highIdx = self._lowIdx
        self._privateSequence = None
    # end def

    def length(self):
        """
        This is the length of a sequence that is immutable by the strand
        """
        return self._highIdx - self._lowIdx + 1

    def modifierType(self):
        return self._mtype
# end class


# Copyright 2011 Autodesk, Inc.  All rights reserved.
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files ( the "Software"), to deal in 
# the Software without restriction, including without limitation the rights to 
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do 
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all 
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
# SOFTWARE.
#
# http://www.opensource.org/licenses/mit-license.php

"""
predecoratornode.py
Created by Alex Tessier on 2011-08-25
A Maya Node for creating a Staple Modifier indicator on top of a Helix Shape 
"""
import sys
import math
import maya.OpenMaya as OpenMaya
import maya.OpenMayaMPx as OpenMayaMPx


nodeName = "spPreDecoratorNode"
id = OpenMaya.MTypeId( 0x00117703 )


gons = [
    2,10,1,
    1,11,2,
    1,8,7,
    1,7,11,
    1,10,8,
    5,2,6,
    10,2,5,
    2,11,6,
    4,9,3,
    3,12,4,
    5,6,3,
    3,9,5,
    6,12,3,
    7,8,4,
    4,12,7,
    4,8,9,
    5,9,10,
    6,11,12,
    7,12,11,
    8,10,9 ]

class PreDecoratorNode( OpenMayaMPx.MPxNode):
    outputMesh = OpenMaya.MObject()
    radiusAttr = OpenMaya.MObject()
    pointArray = OpenMaya.MObject()
    vtx = []

    def __init__( self):
        OpenMayaMPx.MPxNode.__init__( self)

    def compute( self, plug, data):
        radiusData = data.inputValue( PreDecoratorNode.radiusAttr )
        radius = radiusData.asDouble()

        # make the meshDataObj for the plug
        fnMeshData = OpenMaya.MFnMeshData()
        outMeshDataObj = fnMeshData.create()

        # create the mesh and assign to the data
        self.createMesh( radius, outMeshDataObj )
        handle = data.outputValue( PreDecoratorNode.outputMesh )
        handle.setMObject( outMeshDataObj )
        handle.setClean()

        data.setClean( plug )

    def createMesh( self, radius, outputMeshDataObj ):
        self.vtx = []
        self.createPoints( radius )
        num_verts = 12
        num_faces  = 20
        edges_per_face = 3

        pointArray = OpenMaya.MFloatPointArray()
        for i in range( 0, num_verts ):
            pointArray.append( self.vtx[i] ) # can we skip this copy?

        nFaceConnects = num_faces * edges_per_face;
        nEdges = nFaceConnects / 2

        faceCounts = OpenMaya.MIntArray()
        for i in range( 0, num_faces ):
            faceCounts.append( edges_per_face )

        faceConnects = OpenMaya.MIntArray()
        for i in range( 0, num_faces*edges_per_face ):
            faceConnects.append( gons[i] - 1 )

        outputMesh = OpenMaya.MFnMesh()
        newTransform = outputMesh.create( num_verts, num_faces, pointArray, \
                                          faceCounts, faceConnects, \
                                          outputMeshDataObj )
        outputMesh.updateSurface()

    def FILL(  self, x, y, z ):
        self.vtx.append( OpenMaya.MFloatPoint(  x, y, z ) )

    def createPoints( self, radius ):
        a = math.sqrt(  (  1.0 - math.sqrt(  .2 ) ) / 2.0 ) * radius
        b = math.sqrt(  (  1.0 + math.sqrt(  .2 ) ) / 2.0 ) * radius
        z = 0.0

        self.FILL(  b,  a,  z )
        self.FILL(  b, -a,  z )
        self.FILL( -b, -a,  z )
        self.FILL( -b,  a,  z )

        self.FILL(  0, -b, -a )
        self.FILL(  0, -b,  a )
        self.FILL(  0,  b,  a )
        self.FILL(  0,  b, -a )

        self.FILL( -a,  0, -b )
        self.FILL(  a,  0, -b )
        self.FILL(  a,  0,  b )
        self.FILL( -a,  0 , b )

def nodeCreator( ):
    return OpenMayaMPx.asMPxPtr( PreDecoratorNode() )

def nodeInitialize( ):
    typedAttr = OpenMaya.MFnTypedAttribute()
    PreDecoratorNode.outputMesh = typedAttr.create( "outputMesh",
                                                      "om",
                                                       OpenMaya.MFnData.kMesh )

    nAttr = OpenMaya.MFnNumericAttribute()

    PreDecoratorNode.radiusAttr = nAttr.create( 'radius',
                                                  'r',
                                                  OpenMaya.MFnNumericData.kDouble,
                                                  1.0 )

    nAttr.setStorable( True )
    nAttr.setWritable( True )

    PreDecoratorNode.addAttribute( PreDecoratorNode.outputMesh )
    PreDecoratorNode.addAttribute( PreDecoratorNode.radiusAttr )


    PreDecoratorNode.attributeAffects(
        PreDecoratorNode.radiusAttr,
        PreDecoratorNode.outputMesh)


def initializePlugin( obj ):
    plugin = OpenMayaMPx.MFnPlugin( obj)

    try:
        plugin.registerNode( nodeName, id, nodeCreator, nodeInitialize)
    except:
        sys.stderr.write( "Failed to register node %s\n" % nodeName)
        raise

def uninitializePlugin( obj):
    plugin = OpenMayaMPx.MFnPlugin( obj)

    try:
        plugin.deregisterNode( id)
    except:
        sys.stderr.write( "Failed to deregister node %s\n" % nodeName)
        raise



# Form implementation generated from reading ui file 'mainwindow/mainwindow.ui'
#
# Created by: PyQt6 UI code generator 6.1.1
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets
from os import fspath
from pathlib import Path

import cadnano2.util as util

CURRENT_DIRECTORY = Path(__file__).resolve().parent

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        QtCore.QDir.addSearchPath('icons', fspath(CURRENT_DIRECTORY / "images"))
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(1075, 792)
        MainWindow.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.NoContextMenu)
        MainWindow.setStatusTip("")
        MainWindow.setIconSize(QtCore.QSize(32, 32))
        MainWindow.setDockNestingEnabled(True)
        MainWindow.setDockOptions(QtWidgets.QMainWindow.DockOption.AllowNestedDocks|QtWidgets.QMainWindow.DockOption.AllowTabbedDocks|QtWidgets.QMainWindow.DockOption.AnimatedDocks)
        MainWindow.setUnifiedTitleAndToolBarOnMac(True)
        MainWindow.setAcceptDrops(True)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setAutoFillBackground(False)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setHorizontalSpacing(6)
        self.gridLayout.setObjectName("gridLayout")
        self.splitter = QtWidgets.QSplitter(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.splitter.sizePolicy().hasHeightForWidth())
        self.splitter.setSizePolicy(sizePolicy)
        self.splitter.setFrameShape(QtWidgets.QFrame.Shape.NoFrame)
        self.splitter.setFrameShadow(QtWidgets.QFrame.Shadow.Plain)
        self.splitter.setLineWidth(0)
        self.splitter.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.splitter.setOpaqueResize(True)
        self.splitter.setHandleWidth(7)
        self.splitter.setObjectName("splitter")
        self.sliceGraphicsView = CustomQGraphicsView(self.splitter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.sliceGraphicsView.sizePolicy().hasHeightForWidth())
        self.sliceGraphicsView.setSizePolicy(sizePolicy)
        self.sliceGraphicsView.setMinimumSize(QtCore.QSize(0, 0))
        self.sliceGraphicsView.setBaseSize(QtCore.QSize(480, 0))
        self.sliceGraphicsView.setMouseTracking(True)
        self.sliceGraphicsView.setFocusPolicy(QtCore.Qt.FocusPolicy.WheelFocus)
        self.sliceGraphicsView.setFrameShadow(QtWidgets.QFrame.Shadow.Plain)
        self.sliceGraphicsView.setLineWidth(0)
        self.sliceGraphicsView.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.sliceGraphicsView.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.sliceGraphicsView.setRenderHints(QtGui.QPainter.RenderHint.Antialiasing|QtGui.QPainter.RenderHint.TextAntialiasing)
        self.sliceGraphicsView.setDragMode(QtWidgets.QGraphicsView.DragMode.NoDrag)
        self.sliceGraphicsView.setTransformationAnchor(QtWidgets.QGraphicsView.ViewportAnchor.AnchorUnderMouse)
        self.sliceGraphicsView.setResizeAnchor(QtWidgets.QGraphicsView.ViewportAnchor.AnchorUnderMouse)
        self.sliceGraphicsView.setObjectName("sliceGraphicsView")
        self.pathGraphicsView = CustomQGraphicsView(self.splitter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pathGraphicsView.sizePolicy().hasHeightForWidth())
        self.pathGraphicsView.setSizePolicy(sizePolicy)
        self.pathGraphicsView.setMinimumSize(QtCore.QSize(0, 0))
        self.pathGraphicsView.setMouseTracking(True)
        self.pathGraphicsView.setFocusPolicy(QtCore.Qt.FocusPolicy.WheelFocus)
        self.pathGraphicsView.setFrameShadow(QtWidgets.QFrame.Shadow.Plain)
        self.pathGraphicsView.setLineWidth(0)
        self.pathGraphicsView.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.pathGraphicsView.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.pathGraphicsView.setRenderHints(QtGui.QPainter.RenderHint.Antialiasing|QtGui.QPainter.RenderHint.TextAntialiasing)
        self.pathGraphicsView.setTransformationAnchor(QtWidgets.QGraphicsView.ViewportAnchor.AnchorUnderMouse)
        self.pathGraphicsView.setObjectName("pathGraphicsView")
        self.gridLayout.addWidget(self.splitter, 0, 0, 1, 1)
        MainWindow.setCentralWidget(self.centralwidget)
        self.topToolBar = QtWidgets.QToolBar(MainWindow)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Ignored, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.topToolBar.sizePolicy().hasHeightForWidth())
        self.topToolBar.setSizePolicy(sizePolicy)
        self.topToolBar.setBaseSize(QtCore.QSize(550, 0))
        self.topToolBar.setIconSize(QtCore.QSize(24, 24))
        self.topToolBar.setToolButtonStyle(QtCore.Qt.ToolButtonStyle.ToolButtonTextUnderIcon)
        self.topToolBar.setObjectName("topToolBar")
        MainWindow.addToolBar(QtCore.Qt.ToolBarArea.TopToolBarArea, self.topToolBar)
        self.rightToolBar = QtWidgets.QToolBar(MainWindow)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.rightToolBar.sizePolicy().hasHeightForWidth())
        self.rightToolBar.setSizePolicy(sizePolicy)
        self.rightToolBar.setMinimumSize(QtCore.QSize(0, 0))
        self.rightToolBar.setLayoutDirection(QtCore.Qt.LayoutDirection.LeftToRight)
        self.rightToolBar.setAllowedAreas(QtCore.Qt.ToolBarArea.LeftToolBarArea|QtCore.Qt.ToolBarArea.RightToolBarArea|QtCore.Qt.ToolBarArea.TopToolBarArea)
        self.rightToolBar.setToolButtonStyle(QtCore.Qt.ToolButtonStyle.ToolButtonTextUnderIcon)
        self.rightToolBar.setObjectName("rightToolBar")
        with util.suppress_stdout_stderr():
            MainWindow.addToolBar(QtCore.Qt.ToolBarArea.RightToolBarArea, self.rightToolBar)
        self.leftToolBar = QtWidgets.QToolBar(MainWindow)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.leftToolBar.sizePolicy().hasHeightForWidth())
        self.leftToolBar.setSizePolicy(sizePolicy)
        self.leftToolBar.setMinimumSize(QtCore.QSize(0, 0))
        self.leftToolBar.setAllowedAreas(QtCore.Qt.ToolBarArea.LeftToolBarArea|QtCore.Qt.ToolBarArea.RightToolBarArea|QtCore.Qt.ToolBarArea.TopToolBarArea)
        self.leftToolBar.setToolButtonStyle(QtCore.Qt.ToolButtonStyle.ToolButtonTextUnderIcon)
        self.leftToolBar.setObjectName("leftToolBar")
        with util.suppress_stdout_stderr():
            MainWindow.addToolBar(QtCore.Qt.ToolBarArea.LeftToolBarArea, self.leftToolBar)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1075, 22))
        self.menubar.setObjectName("menubar")
        self.menuFile = QtWidgets.QMenu(self.menubar)
        self.menuFile.setObjectName("menuFile")
        self.menuEdit = QtWidgets.QMenu(self.menubar)
        self.menuEdit.setObjectName("menuEdit")
        self.menuPlugins = QtWidgets.QMenu(self.menubar)
        self.menuPlugins.setObjectName("menuPlugins")
        MainWindow.setMenuBar(self.menubar)
        self.actionNew = QtGui.QAction(MainWindow)
        icon = QtGui.QIcon('icons:data-new_32.png')
        self.actionNew.setIcon(icon)
        self.actionNew.setObjectName("actionNew")
        self.actionDrop = QtGui.QAction(MainWindow)
        self.actionDrop.setObjectName("actionDrop")
        self.actionOpen = QtGui.QAction(MainWindow)
        icon1 = QtGui.QIcon('icons:data-open_32.png')
        self.actionOpen.setIcon(icon1)
        self.actionOpen.setObjectName("actionOpen")
        self.actionClose = QtGui.QAction(MainWindow)
        self.actionClose.setObjectName("actionClose")
        self.actionSave = QtGui.QAction(MainWindow)
        icon2 = QtGui.QIcon('icons:data-save_32.png')
        self.actionSave.setIcon(icon2)
        self.actionSave.setObjectName("actionSave")
        self.actionSave_As = QtGui.QAction(MainWindow)
        self.actionSave_As.setIcon(icon2)
        self.actionSave_As.setObjectName("actionSave_As")
        self.actionSave_a_Copy = QtGui.QAction(MainWindow)
        self.actionSave_a_Copy.setObjectName("actionSave_a_Copy")
        self.actionPrint = QtGui.QAction(MainWindow)
        self.actionPrint.setObjectName("actionPrint")
        self.actionSVG = QtGui.QAction(MainWindow)
        icon3 = QtGui.QIcon('icons:data-svg.png')
        self.actionSVG.setIcon(icon3)
        self.actionSVG.setObjectName("actionSVG")
        self.actionX3D = QtGui.QAction(MainWindow)
        self.actionX3D.setObjectName("actionX3D")
        self.actionCut = QtGui.QAction(MainWindow)
        self.actionCut.setObjectName("actionCut")
        self.actionCopy = QtGui.QAction(MainWindow)
        self.actionCopy.setObjectName("actionCopy")
        self.actionPaste = QtGui.QAction(MainWindow)
        self.actionPaste.setObjectName("actionPaste")
        self.actionSelect_All = QtGui.QAction(MainWindow)
        self.actionSelect_All.setObjectName("actionSelect_All")
        self.actionNewHoneycombPart = QtGui.QAction(MainWindow)
        icon4 = QtGui.QIcon('icons:part-new-honeycomb.png')
        self.actionNewHoneycombPart.setIcon(icon4)
        self.actionNewHoneycombPart.setObjectName("actionNewHoneycombPart")
        self.actionPathBreak = QtGui.QAction(MainWindow)
        self.actionPathBreak.setCheckable(True)
        self.actionPathBreak.setChecked(False)
        icon5 = QtGui.QIcon('icons:path-break_48.png')
        self.actionPathBreak.setIcon(icon5)
        self.actionPathBreak.setObjectName("actionPathBreak")
        self.actionPathSelect = QtGui.QAction(MainWindow)
        self.actionPathSelect.setCheckable(True)
        icon6 = QtGui.QIcon('icons:path-edit_32.png')
        self.actionPathSelect.setIcon(icon6)
        self.actionPathSelect.setObjectName("actionPathSelect")
        self.actionSliceFirst = QtGui.QAction(MainWindow)
        icon7 = QtGui.QIcon('icons:slice-go-first.png')
        self.actionSliceFirst.setIcon(icon7)
        self.actionSliceFirst.setObjectName("actionSliceFirst")
        self.actionSliceLast = QtGui.QAction(MainWindow)
        icon8 = QtGui.QIcon('icons:slice-go-last.png')
        self.actionSliceLast.setIcon(icon8)
        self.actionSliceLast.setObjectName("actionSliceLast")
        self.actionPathErase = QtGui.QAction(MainWindow)
        self.actionPathErase.setCheckable(True)
        icon9 = QtGui.QIcon('icons:path-erase_48x48.png')
        self.actionPathErase.setIcon(icon9)
        self.actionPathErase.setObjectName("actionPathErase")
        self.actionPathPencil = QtGui.QAction(MainWindow)
        self.actionPathPencil.setCheckable(True)
        icon10 = QtGui.QIcon('icons:path-force-xover.png')
        self.actionPathPencil.setIcon(icon10)
        self.actionPathPencil.setObjectName("actionPathPencil")
        self.actionPathInsertion = QtGui.QAction(MainWindow)
        self.actionPathInsertion.setCheckable(True)
        icon11 = QtGui.QIcon('icons:path-insert_48x48.png')
        self.actionPathInsertion.setIcon(icon11)
        self.actionPathInsertion.setObjectName("actionPathInsertion")
        self.actionNewSquarePart = QtGui.QAction(MainWindow)
        icon12 = QtGui.QIcon('icons:part-new-square.png')
        self.actionNewSquarePart.setIcon(icon12)
        self.actionNewSquarePart.setObjectName("actionNewSquarePart")
        self.actionPathSkip = QtGui.QAction(MainWindow)
        self.actionPathSkip.setCheckable(True)
        icon13 = QtGui.QIcon('icons:path-skip.png')
        self.actionPathSkip.setIcon(icon13)
        self.actionPathSkip.setObjectName("actionPathSkip")
        self.actionRenumber = QtGui.QAction(MainWindow)
        icon14 = QtGui.QIcon('icons:slice-renumber.png')
        self.actionRenumber.setIcon(icon14)
        self.actionRenumber.setObjectName("actionRenumber")
        self.actionPathPaint = QtGui.QAction(MainWindow)
        self.actionPathPaint.setCheckable(True)
        icon15 = QtGui.QIcon('icons:path-paint_48.png')
        self.actionPathPaint.setIcon(icon15)
        self.actionPathPaint.setObjectName("actionPathPaint")
        self.actionPathAddSeq = QtGui.QAction(MainWindow)
        self.actionPathAddSeq.setCheckable(True)
        icon16 = QtGui.QIcon('icons:path-addseq.png')
        self.actionPathAddSeq.setIcon(icon16)
        self.actionPathAddSeq.setObjectName("actionPathAddSeq")
        self.actionExportStaples = QtGui.QAction(MainWindow)
        icon17 = QtGui.QIcon('icons:export-staples.png')
        self.actionExportStaples.setIcon(icon17)
        self.actionExportStaples.setObjectName("actionExportStaples")
        self.actionPreferences = QtGui.QAction(MainWindow)
        self.actionPreferences.setObjectName("actionPreferences")
        self.actionModify = QtGui.QAction(MainWindow)
        self.actionModify.setCheckable(True)
        icon18 = QtGui.QIcon('icons:path-modify.png')
        self.actionModify.setIcon(icon18)
        self.actionModify.setObjectName("actionModify")
        self.actionCadnanoWebsite = QtGui.QAction(MainWindow)
        self.actionCadnanoWebsite.setObjectName("actionCadnanoWebsite")
        self.actionFeedback = QtGui.QAction(MainWindow)
        self.actionFeedback.setObjectName("actionFeedback")
        self.actionFilterPart = QtGui.QAction(MainWindow)
        self.actionFilterPart.setCheckable(True)
        icon19 = QtGui.QIcon('icons:part-filter-part.png')
        self.actionFilterPart.setIcon(icon19)
        self.actionFilterPart.setObjectName("actionFilterPart")
        self.actionFilterEndpoint = QtGui.QAction(MainWindow)
        self.actionFilterEndpoint.setCheckable(True)
        self.actionFilterEndpoint.setChecked(True)
        icon20 = QtGui.QIcon('icons:part-filter-endpoint.png')
        self.actionFilterEndpoint.setIcon(icon20)
        self.actionFilterEndpoint.setObjectName("actionFilterEndpoint")
        self.actionFilterXover = QtGui.QAction(MainWindow)
        self.actionFilterXover.setCheckable(True)
        self.actionFilterXover.setChecked(True)
        icon21 = QtGui.QIcon('icons:part-filter-xover.png')
        self.actionFilterXover.setIcon(icon21)
        self.actionFilterXover.setText("")
        self.actionFilterXover.setObjectName("actionFilterXover")
        self.actionFiltersLabel = QtGui.QAction(MainWindow)
        self.actionFiltersLabel.setEnabled(False)
        self.actionFiltersLabel.setObjectName("actionFiltersLabel")
        self.actionFilterStrand = QtGui.QAction(MainWindow)
        self.actionFilterStrand.setCheckable(True)
        icon22 = QtGui.QIcon('icons:part-filter-strand.png')
        self.actionFilterStrand.setIcon(icon22)
        self.actionFilterStrand.setObjectName("actionFilterStrand")
        self.actionFilterHandle = QtGui.QAction(MainWindow)
        self.actionFilterHandle.setCheckable(True)
        self.actionFilterHandle.setChecked(False)
        icon23 = QtGui.QIcon('icons:part-filter-handle.png')
        self.actionFilterHandle.setIcon(icon23)
        self.actionFilterHandle.setObjectName("actionFilterHandle")
        self.actionFilterScaf = QtGui.QAction(MainWindow)
        self.actionFilterScaf.setCheckable(True)
        self.actionFilterScaf.setChecked(True)
        icon24 = QtGui.QIcon('icons:part-filter-scaf.png')
        self.actionFilterScaf.setIcon(icon24)
        self.actionFilterScaf.setObjectName("actionFilterScaf")
        self.actionFilterStap = QtGui.QAction(MainWindow)
        self.actionFilterStap.setCheckable(True)
        self.actionFilterStap.setChecked(True)
        icon25 = QtGui.QIcon('icons:part-filter-stap.png')
        self.actionFilterStap.setIcon(icon25)
        self.actionFilterStap.setObjectName("actionFilterStap")
        self.actionAbout = QtGui.QAction(MainWindow)
        self.actionAbout.setObjectName("actionAbout")
        self.actionAutoStaple = QtGui.QAction(MainWindow)
        icon26 = QtGui.QIcon('icons:path-staple.png')
        self.actionAutoStaple.setIcon(icon26)
        self.actionAutoStaple.setObjectName("actionAutoStaple")
        self.topToolBar.addAction(self.actionNew)
        self.topToolBar.addAction(self.actionOpen)
        self.topToolBar.addAction(self.actionSave)
        self.topToolBar.addAction(self.actionSVG)
        self.topToolBar.addAction(self.actionExportStaples)
        self.topToolBar.addSeparator()
        self.topToolBar.addAction(self.actionNewHoneycombPart)
        self.topToolBar.addAction(self.actionNewSquarePart)
        self.topToolBar.addSeparator()
        self.topToolBar.addAction(self.actionAutoStaple)
        self.topToolBar.addAction(self.actionFiltersLabel)
        self.topToolBar.addAction(self.actionFilterScaf)
        self.topToolBar.addAction(self.actionFilterStap)
        self.topToolBar.addAction(self.actionFilterHandle)
        self.topToolBar.addAction(self.actionFilterEndpoint)
        self.topToolBar.addAction(self.actionFilterXover)
        self.topToolBar.addAction(self.actionFilterStrand)
        self.rightToolBar.addAction(self.actionPathSelect)
        self.rightToolBar.addAction(self.actionPathPencil)
        self.rightToolBar.addAction(self.actionPathBreak)
        self.rightToolBar.addAction(self.actionPathInsertion)
        self.rightToolBar.addAction(self.actionPathSkip)
        self.rightToolBar.addAction(self.actionPathPaint)
        self.rightToolBar.addAction(self.actionPathAddSeq)
        self.leftToolBar.addAction(self.actionSliceFirst)
        self.leftToolBar.addAction(self.actionSliceLast)
        self.leftToolBar.addAction(self.actionRenumber)
        self.menuFile.addAction(self.actionAbout)
        self.menuFile.addAction(self.actionPreferences)
        self.menuFile.addAction(self.actionNew)
        self.menuFile.addAction(self.actionOpen)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionClose)
        self.menuFile.addAction(self.actionSave)
        self.menuFile.addAction(self.actionSave_As)
        self.menuEdit.addAction(self.actionModify)
        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuEdit.menuAction())
        self.menubar.addAction(self.menuPlugins.menuAction())

        self.retranslateUi(MainWindow)
        self.actionClose.triggered.connect(MainWindow.close)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "cadnano"))
        self.topToolBar.setWindowTitle(_translate("MainWindow", "Main Toolbar"))
        self.rightToolBar.setWindowTitle(_translate("MainWindow", "Path Tools"))
        self.leftToolBar.setWindowTitle(_translate("MainWindow", "Slice Tools"))
        self.menuFile.setTitle(_translate("MainWindow", "File"))
        self.menuEdit.setTitle(_translate("MainWindow", "Edit"))
        self.menuPlugins.setTitle(_translate("MainWindow", "Plugins"))
        self.actionNew.setText(_translate("MainWindow", "New..."))
        self.actionNew.setShortcut(_translate("MainWindow", "Ctrl+N"))
        self.actionOpen.setText(_translate("MainWindow", "Open..."))
        self.actionOpen.setShortcut(_translate("MainWindow", "Ctrl+O"))
        self.actionClose.setText(_translate("MainWindow", "Close"))
        self.actionClose.setShortcut(_translate("MainWindow", "Ctrl+W"))
        self.actionSave.setText(_translate("MainWindow", "Save"))
        self.actionSave.setShortcut(_translate("MainWindow", "Ctrl+S"))
        self.actionSave_As.setText(_translate("MainWindow", "Save As..."))
        self.actionSave_As.setShortcut(_translate("MainWindow", "Ctrl+Shift+S"))
        self.actionSave_a_Copy.setText(_translate("MainWindow", "Save a Copy..."))
        self.actionPrint.setText(_translate("MainWindow", "Print..."))
        self.actionSVG.setText(_translate("MainWindow", "SVG"))
        self.actionX3D.setText(_translate("MainWindow", "X3D"))
        self.actionCut.setText(_translate("MainWindow", "Cut"))
        self.actionCut.setShortcut(_translate("MainWindow", "Ctrl+X"))
        self.actionCopy.setText(_translate("MainWindow", "Copy"))
        self.actionCopy.setShortcut(_translate("MainWindow", "Ctrl+C"))
        self.actionPaste.setText(_translate("MainWindow", "Paste"))
        self.actionPaste.setShortcut(_translate("MainWindow", "Ctrl+V"))
        self.actionSelect_All.setText(_translate("MainWindow", "Select All"))
        self.actionSelect_All.setShortcut(_translate("MainWindow", "Ctrl+A"))
        self.actionNewHoneycombPart.setText(_translate("MainWindow", "Honeycomb"))
        self.actionNewHoneycombPart.setToolTip(_translate("MainWindow", "Click to add new part with honeycomb lattice"))
        self.actionPathBreak.setText(_translate("MainWindow", "Break Tool"))
        self.actionPathBreak.setIconText(_translate("MainWindow", "Break"))
        self.actionPathBreak.setToolTip(_translate("MainWindow", "(B)reak Tool"))
        self.actionPathBreak.setShortcut(_translate("MainWindow", "B"))
        self.actionPathSelect.setText(_translate("MainWindow", "Select"))
        self.actionPathSelect.setIconText(_translate("MainWindow", "Select"))
        self.actionPathSelect.setToolTip(_translate("MainWindow", "Select Tool (v)"))
        self.actionPathSelect.setShortcut(_translate("MainWindow", "V"))
        self.actionSliceFirst.setText(_translate("MainWindow", "First"))
        self.actionSliceFirst.setToolTip(_translate("MainWindow", "Move the slice bar to the first position."))
        self.actionSliceLast.setText(_translate("MainWindow", "Last"))
        self.actionSliceLast.setToolTip(_translate("MainWindow", "Move the slice bar to the last position."))
        self.actionPathErase.setText(_translate("MainWindow", "Erase"))
        self.actionPathErase.setToolTip(_translate("MainWindow", "(E)rase Tool"))
        self.actionPathPencil.setText(_translate("MainWindow", "Pencil"))
        self.actionPathPencil.setToolTip(_translate("MainWindow", "Pe(n)cil Tool"))
        self.actionPathPencil.setShortcut(_translate("MainWindow", "N"))
        self.actionPathInsertion.setText(_translate("MainWindow", "Insert"))
        self.actionPathInsertion.setToolTip(_translate("MainWindow", "(I)nsert Tool"))
        self.actionPathInsertion.setShortcut(_translate("MainWindow", "I"))
        self.actionNewSquarePart.setText(_translate("MainWindow", "Square"))
        self.actionNewSquarePart.setToolTip(_translate("MainWindow", "Click to add new part with square lattice"))
        self.actionPathSkip.setText(_translate("MainWindow", "Skip"))
        self.actionPathSkip.setToolTip(_translate("MainWindow", "(S)kip Tool"))
        self.actionPathSkip.setShortcut(_translate("MainWindow", "S"))
        self.actionRenumber.setText(_translate("MainWindow", "Rnum"))
        self.actionRenumber.setToolTip(_translate("MainWindow", "Renumber Slice helices according to helix ordering in Path panel."))
        self.actionPathPaint.setText(_translate("MainWindow", "Paint"))
        self.actionPathPaint.setToolTip(_translate("MainWindow", "(P)aint Tool"))
        self.actionPathPaint.setShortcut(_translate("MainWindow", "P"))
        self.actionPathAddSeq.setText(_translate("MainWindow", "Seq"))
        self.actionPathAddSeq.setToolTip(_translate("MainWindow", "(A)dd Sequence Tool"))
        self.actionPathAddSeq.setShortcut(_translate("MainWindow", "A"))
        self.actionExportStaples.setText(_translate("MainWindow", "Export"))
        self.actionExportStaples.setToolTip(_translate("MainWindow", "export oligos as *.CSV"))
        self.actionPreferences.setText(_translate("MainWindow", "Preferences"))
        self.actionPreferences.setShortcut(_translate("MainWindow", "Ctrl+,"))
        self.actionModify.setText(_translate("MainWindow", "Modify mode"))
        self.actionModify.setToolTip(_translate("MainWindow", "Modify mode"))
        self.actionCadnanoWebsite.setText(_translate("MainWindow", "cadnano Website"))
        self.actionFeedback.setText(_translate("MainWindow", "Feedback"))
        self.actionFilterPart.setText(_translate("MainWindow", "Parts"))
        self.actionFilterPart.setToolTip(_translate("MainWindow", "Part-selection mode"))
        self.actionFilterEndpoint.setToolTip(_translate("MainWindow", "(E)ndpoints"))
        self.actionFilterEndpoint.setShortcut(_translate("MainWindow", "E"))
        self.actionFilterXover.setToolTip(_translate("MainWindow", "(X)overs"))
        self.actionFilterXover.setShortcut(_translate("MainWindow", "X"))
        self.actionFiltersLabel.setText(_translate("MainWindow", "Selectable:"))
        self.actionFiltersLabel.setToolTip(_translate("MainWindow", "Selection Filters"))
        self.actionFilterStrand.setToolTip(_translate("MainWindow", "stran(D)s"))
        self.actionFilterStrand.setShortcut(_translate("MainWindow", "D"))
        self.actionFilterHandle.setToolTip(_translate("MainWindow", "(H)andles"))
        self.actionFilterHandle.setShortcut(_translate("MainWindow", "H"))
        self.actionFilterScaf.setToolTip(_translate("MainWindow", "s(C)affold"))
        self.actionFilterScaf.setShortcut(_translate("MainWindow", "C"))
        self.actionFilterStap.setToolTip(_translate("MainWindow", "s(T)aple"))
        self.actionFilterStap.setShortcut(_translate("MainWindow", "T"))
        self.actionAbout.setText(_translate("MainWindow", "About cadnano"))
        self.actionAutoStaple.setText(_translate("MainWindow", "AutoStaple"))
        self.actionAutoStaple.setToolTip(_translate("MainWindow", "Create staple strands complementary to existing scaffold."))

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        files = [u.toLocalFile() for u in event.mimeData().urls()]
        for fname in files:
            self._dropped_file = fname
            self.actionDrop.trigger()

from cadnano2.views.customqgraphicsview import CustomQGraphicsView
